from __future__ import annotations

from typing import TYPE_CHECKING

from elftools.common.exceptions import ELFError
from elftools.elf.elffile import ELFFile

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from pathlib import Path


def elf_read_dt_needed(fn: Path) -> list[str]:
    needed: list[str] = []
    with fn.open("rb") as f:
        elf = ELFFile(f)
        section = elf.get_section_by_name(".dynamic")
        if section is None:
            msg = f"Could not find soname in {fn}"
            raise ValueError(msg)
        needed.extend(t.needed for t in section.iter_tags() if t.entry.d_tag == "DT_NEEDED")
    return needed


def elf_file_filter(paths: Iterable[Path]) -> Iterator[tuple[Path, ELFFile]]:
    """Filter through an iterator of filenames and load up only ELF
    files
    """
    for path in paths:
        if path.name.endswith(".py"):
            continue
        else:
            try:
                with path.open("rb") as f:
                    candidate = ELFFile(f)
                    yield path, candidate
            except ELFError:
                # not an elf file
                continue


def elf_find_versioned_symbols(elf: ELFFile) -> Iterator[tuple[str, str]]:
    section = elf.get_section_by_name(".gnu.version_r")

    if section is not None:
        for verneed, verneed_iter in section.iter_versions():
            if verneed.name.startswith("ld-linux") or verneed.name in [
                "ld64.so.2",
                "ld64.so.1",
            ]:
                continue
            for vernaux in verneed_iter:
                yield (verneed.name, vernaux.name)


def elf_find_ucs2_symbols(elf: ELFFile) -> Iterator[str]:
    section = elf.get_section_by_name(".dynsym")
    if section is not None:
        # look for UCS2 symbols that are externally referenced
        for sym in section.iter_symbols():
            if (
                "PyUnicodeUCS2_" in sym.name
                and sym["st_shndx"] == "SHN_UNDEF"
                and sym["st_info"]["type"] == "STT_FUNC"
            ):
                yield sym.name


def elf_references_pyfpe_jbuf(elf: ELFFile) -> bool:
    offending_symbol_names = ("PyFPE_jbuf", "PyFPE_dummy", "PyFPE_counter")
    section = elf.get_section_by_name(".dynsym")
    if section is not None:
        # look for symbols that are externally referenced
        for sym in section.iter_symbols():
            if (
                sym.name in offending_symbol_names
                and sym["st_shndx"] == "SHN_UNDEF"
                and sym["st_info"]["type"] in ("STT_FUNC", "STT_NOTYPE")
            ):
                return True
    return False


def elf_is_python_extension(fn: Path, elf: ELFFile) -> tuple[bool, int | None]:
    modname = fn.name.split(".", 1)[0]
    module_init_f = {
        "init" + modname: 2,
        "PyInit_" + modname: 3,
        "_cffi_pypyinit_" + modname: 2,
    }

    sect = elf.get_section_by_name(".dynsym")
    if sect is None:
        return False, None

    for sym in sect.iter_symbols():
        if (
            sym.name in module_init_f
            and sym["st_shndx"] != "SHN_UNDEF"
            and sym["st_info"]["type"] == "STT_FUNC"
        ):
            return True, module_init_f[sym.name]

    return False, None


def get_undefined_symbols(path: Path) -> set[str]:
    undef_symbols = set()
    with path.open("rb") as f:
        elf = ELFFile(f)
        section = elf.get_section_by_name(".dynsym")
        if section is not None:
            # look for all undef symbols
            # if the symbol is weak don't consider it as undefined, it's "optional"
            for sym in section.iter_symbols():
                if sym["st_shndx"] == "SHN_UNDEF" and sym["st_info"]["bind"] != "STB_WEAK":
                    undef_symbols.add(sym.name)
    return undef_symbols


def filter_undefined_symbols(
    path: Path,
    symbols: dict[str, frozenset[str]],
) -> dict[str, list[str]]:
    if not symbols:
        return {}
    undef_symbols = set("*") | get_undefined_symbols(path)
    result = {}
    for lib, sym_list in symbols.items():
        intersection = sym_list & undef_symbols
        if intersection:
            result[lib] = sorted(intersection)
    return result



from __future__ import annotations

import subprocess

from setuptools import setup

subprocess.check_call(
    (
        "gcc",
        "testpackage/testprogram.c",
        "-lgsl",
        "-lgslcblas",
        "-o",
        "testpackage/testprogram",
    ),
)
subprocess.check_call(
    ("gcc", "testpackage/testprogram_nodeps.c", "-o", "testpackage/testprogram_nodeps"),
)

setup(
    name="testpackage",
    version="0.0.1",
    packages=["testpackage"],
    package_data={"testpackage": ["testprogram", "testprogram_nodeps"]},
    # This places these files at a path like
    # "testpackage-0.0.1.data/scripts/testprogram", which is needed to test
    # rewriting ELF binaries installed into the scripts directory.
    #
    # Note that using scripts=[] doesn't work here since setuptools expects the
    # scripts to be text and tries to decode them using UTF-8.
    data_files=[
        ("../scripts", ["testpackage/testprogram", "testpackage/testprogram_nodeps"]),
    ],
)



from __future__ import annotations

import itertools
import json
import logging
import os
import platform
import shutil
import stat
from pathlib import Path
from subprocess import check_call
from typing import TYPE_CHECKING

from auditwheel.elfutils import elf_read_dt_needed
from auditwheel.hashfile import hashfile
from auditwheel.lddtree import LIBPYTHON_RE
from auditwheel.policy import get_replace_platforms
from auditwheel.sboms import create_sbom_for_wheel
from auditwheel.tools import is_subdir, unique_by_index
from auditwheel.wheeltools import WHEEL_INFO_RE, InWheelCtx, add_platforms

if TYPE_CHECKING:
    from collections.abc import Iterable

    from auditwheel.patcher import ElfPatcher
    from auditwheel.wheel_abi import WheelAbIInfo

logger = logging.getLogger(__name__)


def repair_wheel(
    wheel_abi: WheelAbIInfo,
    wheel_path: Path,
    abis: list[str],
    lib_sdir: str,
    out_dir: Path,
    *,
    update_tags: bool,
    patcher: ElfPatcher,
    strip: bool,
    zip_compression_level: int,
) -> Path | None:
    external_refs_by_fn = wheel_abi.full_external_refs
    # Do not repair a pure wheel, i.e. has no external refs
    if not external_refs_by_fn:
        return None

    soname_map: dict[str, tuple[str, Path]] = {}

    out_dir = out_dir.resolve(strict=True)
    wheel_fname = wheel_path.name

    output_wheel = out_dir / wheel_fname
    with InWheelCtx(wheel_path) as ctx:
        ctx.out_wheel = output_wheel
        ctx.zip_compression_level = zip_compression_level

        match = WHEEL_INFO_RE(wheel_fname)
        if not match:
            msg = f"Failed to parse wheel file name: {wheel_fname}"
            raise ValueError(msg)

        dest_dir = Path(match.group("name") + lib_sdir)
        dist_info_dirs = list(ctx.path.glob("*.dist-info"))
        assert len(dist_info_dirs) == 1, (  # noqa: S101
            "Expected exactly one .dist-info directory, "
            f"found {len(dist_info_dirs)}: {dist_info_dirs}"
        )
        sbom_filepaths: list[Path] = []

        # here, fn is a path to an ELF file (lib or executable) in
        # the wheel, and v['libs'] contains its required libs
        for fn, v in external_refs_by_fn.items():
            ext_libs = v[abis[0]].libs
            replacements: list[tuple[str, str]] = []
            for soname, src_path in ext_libs.items():
                # Handle libpython dependencies by removing them
                if LIBPYTHON_RE.match(soname):
                    logger.warning(
                        "Removing %s dependency from %s. "
                        "Linking with libpython is forbidden for manylinux/musllinux wheels.",
                        soname,
                        str(fn),
                    )
                    patcher.remove_needed(fn, soname)
                    continue

                if src_path is None:
                    msg = (
                        "Cannot repair wheel, because required "
                        f'library "{soname}" could not be located'
                    )
                    raise ValueError(msg)

                if not dest_dir.exists():
                    dest_dir.mkdir()
                sbom_filepaths.append(src_path)
                new_soname, new_path = copylib(src_path, dest_dir, patcher)
                soname_map[soname] = (new_soname, new_path)
                replacements.append((soname, new_soname))
            if replacements:
                patcher.replace_needed(fn, *replacements)

            if len(ext_libs) > 0:
                new_fn = fn
                if _path_is_script(fn):
                    new_fn = _replace_elf_script_with_shim(match.group("name"), fn)
                new_rpath = Path("$ORIGIN") / os.path.relpath(dest_dir, new_fn.parent)
                append_rpath_within_wheel(new_fn, str(new_rpath), ctx.name, patcher)

        # we grafted in a bunch of libraries and modified their sonames, but
        # they may have internal dependencies (DT_NEEDED) on one another, so
        # we need to update those records so each now knows about the new
        # name of the other.
        # we also clear or set RPATH depending on the presence of internal dependencies
        for _, path in soname_map.values():
            needed = elf_read_dt_needed(path)  # TODO perf, we already read those at some point
            replacements = []
            for n in needed:
                if n in soname_map:
                    replacements.append((n, soname_map[n][0]))
            if replacements:
                patcher.set_rpath(path, "$ORIGIN")
                patcher.replace_needed(path, *replacements)
            else:
                patcher.clear_rpath(path)

        if update_tags:
            output_wheel = add_platforms(ctx, abis, get_replace_platforms(abis[0]))

        if strip:
            libs_to_strip = [path for (_, path) in soname_map.values()]
            extensions = external_refs_by_fn.keys()
            strip_symbols(itertools.chain(libs_to_strip, extensions))

        # If we grafted packages with identities we add an SBOM to the wheel.
        # We recalculate the checksum at this point because there can be
        # modifications to libraries during patching.
        sbom_data = create_sbom_for_wheel(
            wheel_fname=output_wheel.name,
            sbom_filepaths=sbom_filepaths,
        )
        if sbom_data:
            sbom_dir = Path(dist_info_dirs[0], "sboms")
            sbom_dir.mkdir(exist_ok=True)
            (sbom_dir / "auditwheel.cdx.json").write_text(json.dumps(sbom_data))

    return output_wheel


def strip_symbols(libraries: Iterable[Path]) -> None:
    for lib in libraries:
        logger.info("Stripping symbols from %s", lib)
        check_call(["strip", "-s", lib])


def copylib(src_path: Path, dest_dir: Path, patcher: ElfPatcher) -> tuple[str, Path]:
    """Graft a shared library from the system into the wheel and update the
    relevant links.

    1) Copy the file from src_path to dest_dir/
    2) Rename the shared object from soname to soname.<unique>
    """
    # Copy the a shared library from the system (src_path) into the wheel

    with src_path.open("rb") as f:
        shorthash = hashfile(f)[:8]

    src_name = src_path.name
    base, ext = src_name.split(".", 1)
    new_soname = f"{base}-{shorthash}.{ext}" if not base.endswith(f"-{shorthash}") else src_name

    dest_path = dest_dir / new_soname
    if dest_path.exists():
        return new_soname, dest_path

    logger.debug("Grafting: %s -> %s", src_path, dest_path)
    shutil.copy2(src_path, dest_path)
    statinfo = dest_path.stat()
    if not statinfo.st_mode & stat.S_IWRITE:
        dest_path.chmod(statinfo.st_mode | stat.S_IWRITE)

    patcher.set_soname(dest_path, new_soname)

    return new_soname, dest_path


def append_rpath_within_wheel(
    lib_name: Path,
    rpath: str,
    wheel_base_dir: Path,
    patcher: ElfPatcher,
) -> None:
    """Add a new rpath entry to a file while preserving as many existing
    rpath entries as possible.

    In order to preserve an rpath entry it must:

    1) Point to a location within wheel_base_dir.
    2) Not be a duplicate of an already-existing rpath entry.
    """
    if not lib_name.is_absolute():
        lib_name = lib_name.absolute()
    lib_dir = lib_name.parent
    if not wheel_base_dir.is_absolute():
        wheel_base_dir = wheel_base_dir.absolute()

    def is_valid_rpath(rpath: str) -> bool:
        return _is_valid_rpath(rpath, lib_dir, wheel_base_dir)

    old_rpaths = patcher.get_rpath(lib_name)
    rpaths = list(filter(is_valid_rpath, old_rpaths.split(":")))
    rpaths = unique_by_index([*rpaths, rpath])
    patcher.set_rpath(lib_name, ":".join(rpaths))


def _is_valid_rpath(rpath: str, lib_dir: Path, wheel_base_dir: Path) -> bool:
    full_rpath_entry = _resolve_rpath_tokens(rpath, lib_dir)
    if not Path(full_rpath_entry).is_absolute():
        logger.debug(
            "rpath entry %s could not be resolved to an absolute path -- discarding it.",
            rpath,
        )
        return False
    if not is_subdir(full_rpath_entry, wheel_base_dir):
        logger.debug("rpath entry %s points outside the wheel -- discarding it.", rpath)
        return False
    logger.debug("Preserved rpath entry %s", rpath)
    return True


def _resolve_rpath_tokens(rpath: str, lib_base_dir: Path) -> str:
    # See https://www.man7.org/linux/man-pages/man8/ld.so.8.html#DESCRIPTION
    system_lib_dir = "lib64" if platform.architecture()[0] == "64bit" else "lib"
    system_processor_type = platform.machine()
    token_replacements = {
        "ORIGIN": str(lib_base_dir),
        "LIB": system_lib_dir,
        "PLATFORM": system_processor_type,
    }
    for token, target in token_replacements.items():
        rpath = rpath.replace(f"${token}", target)  # $TOKEN
        rpath = rpath.replace(f"${{{token}}}", target)  # ${TOKEN}
    return rpath


def _path_is_script(path: Path) -> bool:
    # Looks something like "uWSGI-2.0.21.data/scripts/uwsgi"
    components = path.parts
    return len(components) == 3 and components[0].endswith(".data") and components[1] == "scripts"


def _replace_elf_script_with_shim(package_name: str, orig_path: Path) -> Path:
    """Move an ELF script and replace it with a shim.

    We can't directly rewrite the RPATH of ELF executables in the "scripts"
    directory since scripts aren't installed to a consistent relative path to
    platlib files.

    Instead, we move the executable into a special directory in platlib and put
    a shim script in its place which execs the real executable.

    More context: https://github.com/pypa/auditwheel/issues/340

    Returns the new path of the moved executable.
    """
    scripts_dir = Path(f"{package_name}.scripts")
    scripts_dir.mkdir(exist_ok=True)

    new_path = scripts_dir / orig_path.name
    orig_path.rename(new_path)

    with orig_path.open("w", newline="\n") as f:
        f.write(_script_shim(new_path))
    orig_path.chmod(new_path.stat().st_mode)

    return new_path


def _script_shim(binary_path: Path) -> str:
    return f"""\
#!python
import os
import sys
import sysconfig


if __name__ == "__main__":
    os.execv(
        os.path.join(sysconfig.get_path("platlib"), {binary_path.as_posix()!r}),
        sys.argv,
    )
"""



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pytest

from auditwheel.tmpdirs import InGivenDirectory, InTemporaryDirectory


def test_intemporarydirectory() -> None:
    cwd = Path.cwd()
    with InTemporaryDirectory() as path:
        assert path.is_dir()
        assert path.samefile(Path.cwd())
        assert not path.samefile(cwd)
    assert not path.exists()
    assert cwd.samefile(Path.cwd())


def test_intemporarydirectory_name() -> None:
    tmp_dir = InTemporaryDirectory()
    with tmp_dir as path:
        assert tmp_dir.name == path


def test_ingivendirectory(tmp_path: Path) -> None:
    cwd = Path.cwd()
    expected_path = tmp_path / "foo"
    with InGivenDirectory(expected_path) as path:
        assert path.is_dir()
        assert path.samefile(Path.cwd())
        assert path.samefile(expected_path)
    assert path.exists()
    assert cwd.samefile(Path.cwd())


def test_ingivendirectory_cwd(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.chdir(tmp_path)
    with InGivenDirectory() as path:
        assert path.is_dir()
        assert path.samefile(Path.cwd())
        assert path.samefile(tmp_path)
    assert path.exists()


def test_ingivendirectory_name():
    given_dir = InGivenDirectory()
    with given_dir as path:
        assert given_dir.name == path



from __future__ import annotations

import shutil
import subprocess
from pathlib import Path

import pytest

from auditwheel.architecture import Architecture

HERE = Path(__file__).parent.resolve()


@pytest.mark.parametrize("mode", ["repair", "show"])
def test_non_platform_wheel_pure(mode):
    wheel = HERE / "plumbum-1.6.8-py2.py3-none-any.whl"
    proc = subprocess.run(
        ["auditwheel", mode, str(wheel)],
        stderr=subprocess.PIPE,
        text=True,
        check=False,
    )
    assert proc.returncode == 1
    assert "This does not look like a platform wheel" in proc.stderr
    assert "AttributeError" not in proc.stderr


@pytest.mark.parametrize(
    ("mode", "samefile"),
    [("repair", False), ("repair", True), ("show", False)],
)
def test_non_platform_wheel_pure_allow(mode: str, samefile: bool, tmp_path: Path) -> None:
    wheel = HERE / "plumbum-1.6.8-py2.py3-none-any.whl"
    dest_wheel = tmp_path / wheel.name
    if samefile:
        shutil.copy2(wheel, dest_wheel)
        wheel = dest_wheel
    args = ["auditwheel", mode, "--allow-pure-python-wheel", str(wheel)]
    if mode == "repair":
        args.extend(["-w", str(tmp_path)])
    proc = subprocess.run(
        args,
        stderr=subprocess.PIPE,
        text=True,
        check=True,
    )
    assert "This does not look like a platform wheel" in proc.stderr
    assert "AttributeError" not in proc.stderr
    if mode == "repair":
        assert dest_wheel.is_file()
        assert dest_wheel.read_bytes() == wheel.read_bytes()


def test_non_platform_wheel_pure_allow_multiple(tmp_path: Path) -> None:
    wheel = HERE / "plumbum-1.6.8-py2.py3-none-any.whl"
    src_wheels = tmp_path / "src"
    src_wheels.mkdir()
    src_wheel1 = src_wheels / "plumbum-1.6.8-py2-none-any.whl"
    src_wheel2 = src_wheels / "plumbum-1.6.8-py3-none-any.whl"
    shutil.copy2(wheel, src_wheel1)
    shutil.copy2(wheel, src_wheel2)
    dst_wheels = tmp_path / "dst"
    dst_wheel1 = dst_wheels / src_wheel1.name
    dst_wheel2 = dst_wheels / src_wheel2.name

    args = [
        "auditwheel",
        "repair",
        "-w",
        str(dst_wheels),
        "--allow-pure-python-wheel",
        str(src_wheel1),
        str(src_wheel2),
    ]
    proc = subprocess.run(
        args,
        stderr=subprocess.PIPE,
        text=True,
        check=True,
    )
    assert "This does not look like a platform wheel" in proc.stderr
    assert "AttributeError" not in proc.stderr
    for dst_wheel in [dst_wheel1, dst_wheel2]:
        assert dst_wheel.is_file()
        assert dst_wheel.read_bytes() == wheel.read_bytes()


@pytest.mark.parametrize("mode", ["repair", "show"])
@pytest.mark.parametrize("arch", ["armv5l", "mips64"])
@pytest.mark.parametrize("allow_pure_python", [True, False])
def test_non_platform_wheel_unknown_arch(mode, arch, allow_pure_python, tmp_path):
    wheel_name = f"testsimple-0.0.1-cp313-cp313-linux_{arch}.whl"
    wheel_path = HERE / "arch-wheels" / "glibc" / wheel_name
    wheel_x86_64 = tmp_path / f"{wheel_path.stem}_x86_64.whl"
    wheel_x86_64.symlink_to(wheel_path)
    args = ["auditwheel", mode, str(wheel_x86_64)]
    if allow_pure_python:
        args.append("--allow-pure-python-wheel")
    proc = subprocess.run(
        args,
        stderr=subprocess.PIPE,
        text=True,
        check=False,
    )
    assert proc.returncode == 1
    assert "Invalid binary wheel: no ELF executable or" in proc.stderr
    assert "unknown architecture" in proc.stderr
    assert "AttributeError" not in proc.stderr


@pytest.mark.parametrize("mode", ["repair", "show"])
@pytest.mark.parametrize(
    "arch",
    ["aarch64", "armv7l", "i686", "x86_64", "ppc64le", "s390x"],
)
@pytest.mark.parametrize("allow_pure_python", [True, False])
def test_non_platform_wheel_bad_arch(mode, arch, allow_pure_python, tmp_path):
    host_arch = Architecture.detect().value
    if host_arch == arch:
        pytest.skip("host architecture")
    wheel_name = f"testsimple-0.0.1-cp313-cp313-linux_{arch}.whl"
    wheel_path = HERE / "arch-wheels" / "glibc" / wheel_name
    wheel_host = tmp_path / f"{wheel_path.stem}_{host_arch}.whl"
    wheel_host.symlink_to(wheel_path)
    args = ["auditwheel", mode, str(wheel_host)]
    if allow_pure_python:
        args.append("--allow-pure-python-wheel")
    proc = subprocess.run(
        args,
        stderr=subprocess.PIPE,
        text=True,
        check=False,
    )
    assert proc.returncode == 1
    assert "Invalid binary wheel: no ELF executable or" in proc.stderr
    assert f"{arch} architecture" in proc.stderr
    assert "AttributeError" not in proc.stderr



import os
from pathlib import Path

import pytest

from auditwheel.architecture import Architecture
from auditwheel.lddtree import LIBPYTHON_RE, ld_paths_from_arg, ldd, parse_ld_paths
from auditwheel.libc import Libc
from auditwheel.tools import zip2dir

HERE = Path(__file__).parent.resolve(strict=True)


@pytest.mark.parametrize(
    "soname",
    [
        "libpython3.7m.so.1.0",
        "libpython3.9.so.1.0",
        "libpython3.10.so.1.0",
        "libpython999.999.so.1.0",
    ],
)
def test_libpython_re_match(soname: str) -> None:
    assert LIBPYTHON_RE.match(soname)


@pytest.mark.parametrize(
    "soname",
    [
        "libpython3.7m.soa1.0",
        "libpython3.9.so.1a0",
    ],
)
def test_libpython_re_nomatch(soname: str) -> None:
    assert LIBPYTHON_RE.match(soname) is None


def test_libpython(tmp_path: Path, caplog: pytest.LogCaptureFixture) -> None:
    wheel = (
        HERE / ".." / "integration" / "python_mscl-67.0.1.0-cp313-cp313-manylinux2014_aarch64.whl"
    )
    so = tmp_path / "python_mscl" / "_mscl.so"
    zip2dir(wheel, tmp_path)
    result = ldd(so)
    assert "Skip libpython3.13.so.1.0 resolution" in caplog.text
    assert result.interpreter is None
    assert result.libc == Libc.GLIBC
    assert result.platform.baseline_architecture == Architecture.aarch64
    assert result.platform.extended_architecture is None
    assert result.path is not None
    assert result.realpath.samefile(so)
    assert result.needed == (
        "libpython3.13.so.1.0",
        "libstdc++.so.6",
        "libm.so.6",
        "libgcc_s.so.1",
        "libc.so.6",
        "ld-linux-aarch64.so.1",
    )
    # libpython must be present in dependencies without path
    libpython = result.libraries["libpython3.13.so.1.0"]
    assert libpython.soname == "libpython3.13.so.1.0"
    assert libpython.path is None
    assert libpython.platform is None
    assert libpython.realpath is None
    assert libpython.needed == ()


def test_parse_ld_paths():
    here = str(HERE)
    parent = str(HERE.parent)

    assert parse_ld_paths("") == []
    assert parse_ld_paths(f"{here}") == [here]
    assert parse_ld_paths(f"{parent}") == [parent]

    # Order is preserved.
    assert parse_ld_paths(f"{here}:{parent}") == [here, parent]
    assert parse_ld_paths(f"{parent}:{here}") == [parent, here]

    # `..` references are normalized.
    assert parse_ld_paths(f"{here}/..") == [parent]

    # Duplicate paths are deduplicated.
    assert parse_ld_paths(f"{here}:{here}") == [here]

    # Empty paths are equivalent to $PWD.
    cwd = str(Path.cwd())
    assert parse_ld_paths(":") == [cwd]
    assert parse_ld_paths(f"{here}:") == [here, cwd]
    assert parse_ld_paths(f":{here}") == [cwd, here]

    # Nonexistent paths are ignored.
    assert parse_ld_paths("/nonexistent") == []
    assert parse_ld_paths(f"/nonexistent:{here}") == [here]


@pytest.mark.parametrize("origin", ["$ORIGIN", "${ORIGIN}"])
def test_parse_ld_paths_origin(origin):
    here = str(HERE)
    parent = str(HERE.parent)

    with pytest.raises(ValueError, match=r"can't expand \$ORIGIN without a path"):
        parse_ld_paths(origin)

    assert parse_ld_paths(origin, path=__file__) == [here]
    assert parse_ld_paths(f"{origin}/..", path=__file__) == [parent]

    # Relative paths are made absolute.
    assert parse_ld_paths(origin, path=os.path.relpath(__file__)) == [here]


@pytest.mark.parametrize(
    ("arg", "env", "expected"),
    [
        (None, "", None),
        (None, str(HERE.parent), None),
        ("", "", {"conf": [], "env": [], "interp": []}),
        (str(HERE), "", {"conf": [str(HERE)], "env": [], "interp": []}),
        ("", str(HERE), {"conf": [], "env": [str(HERE)], "interp": []}),
        (
            str(HERE),
            str(HERE.parent),
            {"conf": [str(HERE)], "env": [str(HERE.parent)], "interp": []},
        ),
    ],
)
def test_ld_paths_from_arg(arg, env, expected, monkeypatch):
    monkeypatch.setitem(os.environ, "AUDITWHEEL_LD_LIBRARY_PATH", env)
    assert ld_paths_from_arg(arg) == expected



from __future__ import annotations

import dataclasses
import json
from enum import Enum
from pathlib import PurePath
from typing import Any


class _CustomEncoder(json.JSONEncoder):
    def default(self, value: Any) -> Any:  # noqa: ANN401
        if dataclasses.is_dataclass(value) and not isinstance(value, type):
            as_dict = dataclasses.asdict(value)
            as_dict.pop("policy", None)  # don't dump full policy in logs
            return as_dict
        if isinstance(value, frozenset):
            return sorted(value)
        if isinstance(value, Enum):
            return repr(value)
        if isinstance(value, PurePath):
            return str(value)
        return super().default(value)

    def encode(self, o: Any) -> str:  # noqa: ANN401
        if isinstance(o, dict):
            o = {str(k): v for k, v in o.items()}
        return super().encode(o)


def dumps(obj: Any) -> str:  # noqa: ANN401
    return json.dumps(obj, indent=4, cls=_CustomEncoder)



from __future__ import annotations

import hashlib
import typing
from importlib import metadata
from typing import TYPE_CHECKING
from urllib.parse import quote

from auditwheel._vendor.whichprovides import whichprovides
from auditwheel.wheeltools import WHEEL_INFO_RE

if TYPE_CHECKING:
    from pathlib import Path


def create_sbom_for_wheel(
    wheel_fname: str,
    sbom_filepaths: list[Path],
) -> None | dict[str, typing.Any]:
    # If there aren't any files then we bail.
    if not sbom_filepaths:
        return None

    # Pull the top-level package name and version
    # from the wheel filename. This segment doesn't
    # change even after "repairing", so we don't have
    # to worry about it changing.
    match = WHEEL_INFO_RE(wheel_fname)
    if not match:
        msg = f"Failed to parse wheel file name: {wheel_fname}"
        raise ValueError(msg)
    wheel_name = match.group("name")
    wheel_version = match.group("ver")
    wheel_purl = (
        f"pkg:pypi/{quote(wheel_name, safe='')}@{quote(wheel_version, safe='')}"
        f"?file_name={quote(wheel_fname, safe='')}"
    )

    # Lookup which packages provided libraries.
    # If there aren't any then we don't generate an SBOM.
    sbom_packages = whichprovides([str(f) for f in sbom_filepaths])
    if not sbom_packages:
        return None

    sbom_components: list[dict[str, typing.Any]] = [
        {
            "type": "library",
            "bom-ref": wheel_purl,
            "name": wheel_name,
            "version": wheel_version,
            "purl": wheel_purl,
        },
    ]
    sbom_dependencies = [{"ref": wheel_purl, "dependsOn": []}]
    sbom_data = {
        "bomFormat": "CycloneDX",
        "specVersion": "1.4",
        "version": 1,
        "metadata": {
            "component": sbom_components[0],
            "tools": [
                {"name": "auditwheel", "version": metadata.version("auditwheel")},
            ],
        },
        # These are mutated below through the variables.
        "components": sbom_components,
        "dependencies": sbom_dependencies,
    }

    for filepath, provided_by in sbom_packages.items():
        bom_ref = (
            provided_by.purl + f"#{hashlib.sha256(filepath.encode(errors='ignore')).hexdigest()}"
        )
        sbom_components.append(
            {
                "type": "library",
                "bom-ref": bom_ref,
                "name": provided_by.package_name,
                "version": provided_by.package_version,
                "purl": provided_by.purl,
            },
        )
        sbom_dependencies[0]["dependsOn"].append(bom_ref)  # type: ignore[attr-defined]
        sbom_dependencies.append({"ref": bom_ref})

    return sbom_data



from __future__ import annotations

import dataclasses
import functools
import itertools
import logging
from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar

from auditwheel import json
from auditwheel.elfutils import (
    elf_file_filter,
    elf_find_ucs2_symbols,
    elf_find_versioned_symbols,
    elf_is_python_extension,
    elf_references_pyfpe_jbuf,
)
from auditwheel.error import InvalidLibcError, NonPlatformWheelError
from auditwheel.genericpkgctx import InGenericPkgCtx
from auditwheel.lddtree import DynamicExecutable, DynamicLibrary, ld_paths_from_arg, ldd
from auditwheel.libc import Libc
from auditwheel.policy import ExternalReference, Policy, WheelPolicies

if TYPE_CHECKING:
    from collections.abc import Mapping

    from auditwheel.architecture import Architecture

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class WheelAbIInfo:
    policies: WheelPolicies
    full_external_refs: dict[Path, dict[str, ExternalReference]]
    overall_policy: Policy
    external_refs: dict[str, ExternalReference]
    ref_policy: Policy
    versioned_symbols: dict[str, set[str]]
    sym_policy: Policy
    ucs_policy: Policy
    pyfpe_policy: Policy
    blacklist_policy: Policy
    machine_policy: Policy
    graft_policy: Policy


@dataclass(frozen=True)
class WheelElfData:
    policies: WheelPolicies
    full_elftree: dict[Path, DynamicExecutable]
    full_external_refs: dict[Path, dict[str, ExternalReference]]
    versioned_symbols: dict[str, set[str]]
    uses_ucs2_symbols: bool
    uses_pyfpe_jbuf: bool


def _fixup_elf_trees(
    py_elf_trees: dict[Path, DynamicExecutable],
    nonpy_elftree: dict[Path, DynamicExecutable],
) -> None:
    # Auditwheel assumes that if a dependency in a non-Python ELF tree can be resolved
    # in any other fully resolved ELF tree, then it can be loaded from other
    # non-Python ELF trees as well. Python extension ELF trees are not
    # fixed up; they should always be valid in the first place. We fix up
    # non-Python ELF trees with that assumption in mind.
    resolved_executables = [
        executable
        for executable in itertools.chain(py_elf_trees.values(), nonpy_elftree.values())
        if all(library.realpath is not None for library in executable.libraries.values())
    ]

    def _get_reference(path_to_elf: Path) -> DynamicExecutable | None:
        assert path_to_elf.is_file()  # noqa: S101
        for candidate in resolved_executables:
            for library in candidate.libraries.values():
                assert library.realpath is not None  # noqa: S101
                assert library.realpath.is_file()  # noqa: S101
                if path_to_elf.samefile(library.realpath):
                    return candidate
        return None

    for elf_path, elf_executable in nonpy_elftree.items():
        need_resolution = any(
            elf_library.realpath is None for elf_library in elf_executable.libraries.values()
        )
        if not need_resolution:
            continue
        reference = _get_reference(elf_executable.realpath)
        if reference is None:
            continue
        ref_name = reference.realpath.name
        libraries_new: dict[str, DynamicLibrary] = {}
        soname_all = set(elf_executable.libraries.keys())
        for soname, elf_library in elf_executable.libraries.items():
            if elf_library.realpath is None and soname in reference.libraries:
                libraries_new[soname] = reference.libraries[soname]
                # we need to add newly found dependencies as well
                needed_new = set(libraries_new[soname].needed) - soname_all
                while needed_new:
                    needed_new_copy = needed_new.copy()
                    soname_all |= needed_new_copy
                    for soname_needed in needed_new_copy:
                        if soname_needed in reference.libraries:
                            libraries_new[soname_needed] = reference.libraries[soname_needed]
                            needed_new.update(libraries_new[soname_needed].needed)
                        else:
                            log.warning("%s not found in %s ELF tree", soname_needed, ref_name)
                    needed_new -= soname_all
            else:
                if elf_library.realpath is None:
                    log.warning("%s not found in %s ELF tree", soname, ref_name)
                libraries_new[soname] = elf_library
        nonpy_elftree[elf_path] = dataclasses.replace(elf_executable, libraries=libraries_new)


@functools.lru_cache
def get_wheel_elfdata(
    libc: Libc | None,
    architecture: Architecture | None,
    wheel_fn: Path,
    exclude: frozenset[str],
    args_ldpaths: str | None,
) -> WheelElfData:
    full_elftree: dict[Path, DynamicExecutable] = {}
    nonpy_elftree: dict[Path, DynamicExecutable] = {}
    full_external_refs: dict[Path, dict[str, ExternalReference]] = {}
    versioned_symbols: dict[str, set[str]] = defaultdict(set)
    uses_ucs2_symbols = False
    uses_pyfpe_jbuf = False
    policies: WheelPolicies | None = None

    # Android is cross-compiled, so ldpaths should never be loaded from the build machine.
    if libc == Libc.ANDROID and args_ldpaths is None:
        args_ldpaths = ""

    with InGenericPkgCtx(wheel_fn) as ctx:
        shared_libraries_in_purelib = []
        shared_libraries_with_invalid_machine = []

        platform_wheel = False
        for fn, elf in elf_file_filter(ctx.iter_files()):
            # Check for invalid binary wheel format: no shared library should
            # be found in purelib
            so_name = fn.name

            # If this is in purelib, add it to the list of shared libraries in
            # purelib
            if any(p.name == "purelib" for p in fn.parents):
                shared_libraries_in_purelib.append(so_name)

            # If at least one shared library exists in purelib, this is going
            # to fail and there's no need to do further checks
            if not shared_libraries_in_purelib:
                log.debug("processing: %s", fn)
                elftree = ldd(fn, exclude=exclude, ldpaths=ld_paths_from_arg(args_ldpaths))

                try:
                    elf_arch = elftree.platform.baseline_architecture
                except ValueError:
                    shared_libraries_with_invalid_machine.append(so_name)
                    log.warning("ignoring: %s with unknown architecture", so_name)
                    continue
                if architecture is None:
                    log.info("setting architecture to %s", elf_arch.value)
                    architecture = elf_arch
                elif elf_arch != architecture.baseline:
                    shared_libraries_with_invalid_machine.append(so_name)
                    log.warning("ignoring: %s with %s architecture", so_name, elf_arch)
                    continue

                if elftree.libc is not None:
                    if libc is None:
                        log.info("setting libc to %s", elftree.libc)
                        libc = elftree.libc
                    elif libc != elftree.libc:
                        log.warning("ignoring: %s with %s libc", so_name, elftree.libc)
                        continue

                if policies is None and libc is not None and architecture is not None:
                    policies = WheelPolicies(libc=libc, arch=architecture, wheel_fn=wheel_fn)

                platform_wheel = True

                for key, value in elf_find_versioned_symbols(elf):
                    log.debug("key %s, value %s", key, value)
                    versioned_symbols[key].add(value)

                is_py_ext, py_ver = elf_is_python_extension(fn, elf)

                # If the ELF is a Python extention, we definitely need to
                # include its external dependencies.
                if is_py_ext:
                    if policies is None:
                        assert architecture is not None  # noqa: S101
                        assert libc is None  # noqa: S101
                        msg = f"couldn't detect libc for python extension {fn}"
                        raise InvalidLibcError(msg)
                    full_elftree[fn] = elftree
                    uses_pyfpe_jbuf |= elf_references_pyfpe_jbuf(elf)
                    if py_ver == 2:
                        uses_ucs2_symbols |= any(True for _ in elf_find_ucs2_symbols(elf))
                    full_external_refs[fn] = policies.lddtree_external_references(
                        elftree,
                        ctx.path,
                    )
                else:
                    # If the ELF is not a Python extension, it might be
                    # included in the wheel already because auditwheel repair
                    # vendored it, so we will check whether we should include
                    # its internal references later.
                    nonpy_elftree[fn] = elftree

        # If at least one shared library exists in purelib, raise an error
        if shared_libraries_in_purelib:
            libraries = "\n\t".join(shared_libraries_in_purelib)
            msg = (
                "Invalid binary wheel, found the following shared library/libraries "
                f"in purelib folder:\n\t{libraries}\n"
                "The wheel has to be platlib compliant in order to be repaired by "
                "auditwheel."
            )
            raise RuntimeError(msg)

        if not platform_wheel:
            arch = None if architecture is None else architecture.value
            raise NonPlatformWheelError(arch, shared_libraries_with_invalid_machine)

        _fixup_elf_trees(full_elftree, nonpy_elftree)

        # Get a list of all external libraries needed by ELFs in the wheel.
        needed_libs = {
            lib
            for elf in itertools.chain(full_elftree.values(), nonpy_elftree.values())
            for lib in elf.needed
        }

        if policies is None:
            # we have no python extensions, either we have shared libraries with
            # no dependencies on libc (unlikely) or a statically linked executable
            # let's fallback to the host libc
            assert architecture is not None  # noqa: S101
            assert libc is None  # noqa: S101
            libc = Libc.detect()
            log.warning("couldn't detect wheel libc, defaulting to %s", str(libc))
            policies = WheelPolicies(libc=libc, arch=architecture)

        for fn, elf_tree in nonpy_elftree.items():
            # If a non-pyextension ELF file is not needed by something else
            # inside the wheel, then it was not checked by the logic above and
            # we should walk its elftree.
            if fn.name not in needed_libs:
                full_elftree[fn] = elf_tree

            # Even if a non-pyextension ELF file is not needed, we
            # should include it as an external reference, because
            # it might require additional external libraries.
            full_external_refs[fn] = policies.lddtree_external_references(
                elf_tree,
                ctx.path,
            )

    log.debug("full_elftree:\n%s", json.dumps(full_elftree))
    log.debug(
        "full_external_refs (will be repaired):\n%s",
        json.dumps(full_external_refs),
    )

    return WheelElfData(
        policies,
        full_elftree,
        full_external_refs,
        versioned_symbols,
        uses_ucs2_symbols,
        uses_pyfpe_jbuf,
    )


def get_external_libs(external_refs: dict[str, ExternalReference]) -> dict[Path, str]:
    """Get external library dependencies for all policies excluding the default
    linux policy
    :param external_refs: external references for all policies
    :return: {realpath: soname} e.g.
    {'/path/to/external_ref.so.1.2.3': 'external_ref.so.1'}
    """
    result: dict[Path, str] = {}
    for external_ref in external_refs.values():
        # linux tag (priority 0) has no white-list, do not analyze it
        if external_ref.policy.priority == 0:
            continue
        # go through all libs, retrieving their soname and realpath
        for libname, realpath in external_ref.libs.items():
            if realpath and realpath not in result:
                result[Path(realpath)] = libname
    return result


def get_versioned_symbols(libs: dict[Path, str]) -> dict[str, dict[str, set[str]]]:
    """Get versioned symbols used in libraries
    :param libs: {realpath: soname} dict to search for versioned symbols e.g.
    {'/path/to/external_ref.so.1.2.3': 'external_ref.so.1'}
    :return: {soname: {depname: set([symbol_version])}} e.g.
    {'external_ref.so.1': {'libc.so.6', set(['GLIBC_2.5','GLIBC_2.12'])}}
    """
    result = {}
    for path, elf in elf_file_filter(libs.keys()):
        # {depname: set(symbol_version)}, e.g. {'libc.so.6', set(['GLIBC_2.5','GLIBC_2.12'])}
        elf_versioned_symbols: dict[str, set[str]] = defaultdict(set)
        for key, value in elf_find_versioned_symbols(elf):
            log.debug("path %s, key %s, value %s", path, key, value)
            elf_versioned_symbols[key].add(value)
        result[libs[path]] = elf_versioned_symbols
    return result


def get_symbol_policies(
    policies: WheelPolicies,
    versioned_symbols: dict[str, set[str]],
    external_versioned_symbols: dict[str, dict[str, set[str]]],
    external_refs: dict[str, ExternalReference],
) -> list[tuple[Policy, dict[str, set[str]]]]:
    """Get symbol policies
    Since white-list is different per policy, this function inspects
    versioned_symbol per policy when including external refs
    :param versioned_symbols: versioned symbols for the current wheel
    :param external_versioned_symbols: versioned symbols for external libs
    :param external_refs: external references for all policies
    :return: list of tuples of the form (policy, versioned_symbols),
    e.g. [(<Policy: manylinux...>, {'libc.so.6', set(['GLIBC_2.5'])})]
    """
    result = []
    for external_ref in external_refs.values():
        # skip the linux policy
        if external_ref.policy.priority == 0:
            continue
        policy_symbols = deepcopy(versioned_symbols)
        for soname in external_ref.libs:
            if soname not in external_versioned_symbols:
                continue
            ext_symbols = external_versioned_symbols[soname]
            for k in iter(ext_symbols):
                policy_symbols[k].update(ext_symbols[k])
        # if the white-list policy changed, we don't want to allow highest priority policy
        # than the current one, that is, only restrict to a lower priority policy
        found_policy = min(
            external_ref.policy,
            policies.versioned_symbols_policy(policy_symbols),
        )
        result.append((found_policy, policy_symbols))
    return result


def _get_machine_policy(
    policies: WheelPolicies,
    elftree_by_fn: dict[Path, DynamicExecutable],
    external_so_names: frozenset[str],
) -> Policy:
    """
    Determine the most appropriate machine policy for the wheel based on ELF
    architectures and whitelisted external dependencies.
    The function inspects the extended architecture of each top-level ELF file
    in ``elftree_by_fn`` and of any library dependencies whose SONAME appears in
    ``external_so_names``. If an ELF file requires instructions that are not
    covered by ``policies.architecture``, the overall policy is downgraded to
    reflect that requirement.
    If the offending ELF file corresponds to an external shared object
    (i.e. its SONAME is in ``external_so_names``), the function looks for a
    policy whose whitelist contains that SONAME and uses the highest-priority
    such policy instead of unconditionally falling back to the baseline
    ``policies.linux`` policy. Informational or warning logs are emitted to
    record any such downgrades.
    :param policies: Collection of available wheel policies, including the
        target architecture, the baseline ``linux`` policy, and the highest
        (most permissive) default policy.
    :param elftree_by_fn: Mapping from ELF file paths to their corresponding
        :class:`DynamicExecutable` objects, used to inspect platforms and
        dependency graphs.
    :param external_so_names: Set of SONAMEs that are treated as external
        references and may be whitelisted by individual policies.
    :return: The selected :class:`Policy` after considering all relevant ELF
        files and their dependencies.
    """
    result = policies.highest
    machine_to_check: dict[Path, tuple[str | None, Architecture | None]] = {}
    for fn, dynamic_executable in elftree_by_fn.items():
        if fn in machine_to_check:
            continue
        machine_to_check[fn] = (None, dynamic_executable.platform.extended_architecture)
        for dependency in dynamic_executable.libraries.values():
            if dependency.soname not in external_so_names:
                continue
            if dependency.realpath is None:
                continue
            assert dependency.platform is not None  # noqa: S101
            if dependency.realpath in machine_to_check:
                continue
            machine_to_check[dependency.realpath] = (
                dependency.soname,
                dependency.platform.extended_architecture,
            )

    for fn, (soname, extended_architecture) in machine_to_check.items():
        if extended_architecture is None:
            continue
        if policies.architecture.is_superset(extended_architecture):
            continue
        if soname is not None:
            found_policy = policies.linux
            for policy in policies:
                if soname in policy.whitelist:
                    found_policy = max(found_policy, policy)
            if policies.linux.priority < found_policy.priority:
                log.info(
                    "ELF file %r requires %r instruction set, not in %r, whitelisted in %r",
                    fn,
                    extended_architecture.value,
                    policies.architecture.value,
                    found_policy.name,
                )
                result = min(result, found_policy)
                continue
        log.warning(
            "ELF file %r requires %r instruction set, not in %r",
            fn,
            extended_architecture.value,
            policies.architecture.value,
        )
        result = policies.linux

    return result


def analyze_wheel_abi(
    libc: Libc | None,
    architecture: Architecture | None,
    wheel_fn: Path,
    exclude: frozenset[str],
    *,
    disable_isa_ext_check: bool,
    allow_graft: bool,
    requested_policy_base_name: str | None = None,
    args_ldpaths: str | None = None,
) -> WheelAbIInfo:
    data = get_wheel_elfdata(libc, architecture, wheel_fn, exclude, args_ldpaths)
    policies = data.policies
    elftree_by_fn = data.full_elftree
    external_refs_by_fn = data.full_external_refs
    versioned_symbols = data.versioned_symbols

    external_refs: dict[str, ExternalReference] = {
        p.name: ExternalReference({}, {}, p) for p in policies
    }

    for fn in elftree_by_fn:
        update(external_refs, external_refs_by_fn[fn])

    log.debug("external reference info")
    log.debug(json.dumps(external_refs))

    external_libs = get_external_libs(external_refs)
    external_versioned_symbols = get_versioned_symbols(external_libs)
    symbol_policies = get_symbol_policies(
        policies,
        versioned_symbols,
        external_versioned_symbols,
        external_refs,
    )
    symbol_policy = policies.versioned_symbols_policy(versioned_symbols)

    # let's keep the highest priority policy and
    # corresponding versioned_symbols
    symbol_policy, versioned_symbols = max(
        symbol_policies,
        key=lambda x: x[0],
        default=(symbol_policy, versioned_symbols),
    )

    ref_policy = max(
        (e.policy for e in external_refs.values() if len(e.libs) == 0),
        default=policies.linux,
    )

    blacklist_policy = max(
        (e.policy for e in external_refs.values() if len(e.blacklist) == 0),
        default=policies.linux,
    )

    # don't allow highest priority policies with more grafted libraries
    # than the requested policy
    if requested_policy_base_name is None or requested_policy_base_name == "auto":
        graft_lib_count = min(
            (len(e.libs) for e in external_refs.values() if e.policy != policies.linux),
            default=0,
        )
    else:
        graft_lib_count = min(
            (
                len(e.libs)
                for e in external_refs.values()
                if e.policy.name.startswith(f"{requested_policy_base_name}_")
            ),
            default=0,
        )
    graft_policy = max(
        (e.policy for e in external_refs.values() if len(e.libs) <= graft_lib_count),
        default=policies.linux,
    )

    if disable_isa_ext_check:
        machine_policy = policies.highest
    else:
        machine_policy = _get_machine_policy(
            policies,
            elftree_by_fn,
            frozenset(external_libs.values()),
        )

    ucs_policy = policies.linux if data.uses_ucs2_symbols else policies.highest
    pyfpe_policy = policies.linux if data.uses_pyfpe_jbuf else policies.highest

    overall_policy = min(
        symbol_policy,
        ucs_policy,
        pyfpe_policy,
        blacklist_policy,
        machine_policy,
    )

    if requested_policy_base_name is not None:
        overall_policy = min(overall_policy, graft_policy)

    if not allow_graft:
        overall_policy = min(overall_policy, ref_policy)

    return WheelAbIInfo(
        policies,
        external_refs_by_fn,
        overall_policy,
        external_refs,
        ref_policy,
        versioned_symbols,
        symbol_policy,
        ucs_policy,
        pyfpe_policy,
        blacklist_policy,
        machine_policy,
        graft_policy,
    )


_T = TypeVar("_T", ExternalReference, Path | None)


def update(d: dict[str, _T], u: Mapping[str, _T]) -> None:
    for k, v in u.items():
        if isinstance(v, ExternalReference):
            assert k in d  # noqa: S101
            assert isinstance(d[k], ExternalReference)  # noqa: S101
            assert d[k].policy == v.policy  # noqa: S101
            # update blacklist
            for lib, symbols in v.blacklist.items():
                if lib not in d[k].blacklist:
                    d[k].blacklist[lib] = sorted(symbols)
                else:
                    d[k].blacklist[lib] = sorted(
                        set(d[k].blacklist[lib]) | set(symbols),
                    )
            # libs
            update(d[k].libs, v.libs)
        elif isinstance(v, (Path, type(None))):
            d[k] = v
        else:
            msg = f"can't update {d} with {k}:{v}"
            raise TypeError(msg)



from __future__ import annotations

import io
import json
import logging
import os
import re
import shutil
import sys
import zipfile
from contextlib import contextmanager
from pathlib import Path
from subprocess import CalledProcessError
from typing import TYPE_CHECKING, Any

import docker
import pytest

if TYPE_CHECKING:
    from collections.abc import Generator

    from docker.models.containers import Container
from elftools.elf.elffile import ELFFile

from auditwheel.architecture import Architecture
from auditwheel.libc import Libc
from auditwheel.policy import WheelPolicies

logger = logging.getLogger(__name__)

if "GITHUB_ACTIONS" in os.environ and sys.platform == "darwin":
    pytest.skip("Docker is not available on GitHub Actions macOS runners", allow_module_level=True)

PLATFORM = os.environ.get("AUDITWHEEL_ARCH", Architecture.detect().value)
MANYLINUX2010_IMAGE_ID = f"quay.io/pypa/manylinux2010_{PLATFORM}:latest"
MANYLINUX2014_IMAGE_ID = f"quay.io/pypa/manylinux2014_{PLATFORM}:latest"
MANYLINUX_2_28_IMAGE_ID = f"quay.io/pypa/manylinux_2_28_{PLATFORM}:latest"
MANYLINUX_2_31_IMAGE_ID = f"quay.io/pypa/manylinux_2_31_{PLATFORM}:latest"
MANYLINUX_2_34_IMAGE_ID = f"quay.io/pypa/manylinux_2_34_{PLATFORM}:latest"
MANYLINUX_2_35_IMAGE_ID = f"quay.io/pypa/manylinux_2_35_{PLATFORM}:latest"
MANYLINUX_2_39_IMAGE_ID = f"quay.io/pypa/manylinux_2_39_{PLATFORM}:latest"
if PLATFORM in {"i686", "x86_64"}:
    MANYLINUX_IMAGES = {
        "manylinux_2_12": MANYLINUX2010_IMAGE_ID,
        "manylinux_2_17": MANYLINUX2014_IMAGE_ID,
        "manylinux_2_28": MANYLINUX_2_28_IMAGE_ID,
        "manylinux_2_34": MANYLINUX_2_34_IMAGE_ID,
    }
    POLICY_ALIASES = {
        "manylinux_2_5": ["manylinux1"],
        "manylinux_2_12": ["manylinux2010"],
        "manylinux_2_17": ["manylinux2014"],
    }
elif PLATFORM == "armv7l":
    MANYLINUX_IMAGES = {
        "manylinux_2_31": MANYLINUX_2_31_IMAGE_ID,
        "manylinux_2_35": MANYLINUX_2_35_IMAGE_ID,
    }
    POLICY_ALIASES = {}
elif PLATFORM == "riscv64":
    MANYLINUX_IMAGES = {"manylinux_2_39": MANYLINUX_2_39_IMAGE_ID}
    POLICY_ALIASES = {}
else:
    MANYLINUX_IMAGES = {
        "manylinux_2_17": MANYLINUX2014_IMAGE_ID,
        "manylinux_2_28": MANYLINUX_2_28_IMAGE_ID,
    }
    if os.environ.get("AUDITWHEEL_QEMU", "") != "true":
        MANYLINUX_IMAGES.update({"manylinux_2_34": MANYLINUX_2_34_IMAGE_ID})
        if PLATFORM == "aarch64":
            MANYLINUX_IMAGES.update({"manylinux_2_39": MANYLINUX_2_39_IMAGE_ID})
    POLICY_ALIASES = {
        "manylinux_2_17": ["manylinux2014"],
    }
DOCKER_CONTAINER_NAME = "auditwheel-test-anylinux"
PYTHON_MAJ_MIN = [str(i) for i in sys.version_info[:2]]
PYTHON_ABI_MAJ_MIN = "".join(PYTHON_MAJ_MIN)
PYTHON_ABI = f"cp{PYTHON_ABI_MAJ_MIN}-cp{PYTHON_ABI_MAJ_MIN}"
PYTHON_IMAGE_TAG = ".".join(PYTHON_MAJ_MIN) + ("-rc" if PYTHON_ABI_MAJ_MIN == "314" else "")
MANYLINUX_PYTHON_IMAGE_ID = f"python:{PYTHON_IMAGE_TAG}-slim-trixie"
MUSLLINUX_IMAGES = {
    "musllinux_1_2": f"quay.io/pypa/musllinux_1_2_{PLATFORM}:latest",
}
MUSLLINUX_PYTHON_IMAGE_ID = f"python:{PYTHON_IMAGE_TAG}-alpine"
DEVTOOLSET = {
    "manylinux_2_12": "devtoolset-8",
    "manylinux_2_17": "devtoolset-10",
    "manylinux_2_28": "gcc-toolset-14",
    "manylinux_2_31": "devtoolset-not-present",
    "manylinux_2_34": "gcc-toolset-14",
    "manylinux_2_35": "devtoolset-not-present",
    "manylinux_2_39": "devtoolset-not-present",
    "musllinux_1_2": "devtoolset-not-present",
}
PATH_DIRS = [
    f"/opt/python/{PYTHON_ABI}/bin",
    "/opt/rh/{devtoolset}/root/usr/bin",
    "/usr/local/sbin",
    "/usr/local/bin",
    "/usr/sbin",
    "/usr/bin",
    "/sbin",
    "/bin",
]
PATH = {k: ":".join(PATH_DIRS).format(devtoolset=v) for k, v in DEVTOOLSET.items()}
WHEEL_CACHE_FOLDER = Path.home().joinpath(".cache", "auditwheel_tests")
HERE = Path(__file__).parent.resolve(strict=True)
NUMPY_VERSION_MAP = {
    "310": "1.21.4",
    "311": "1.23.4",
    "312": "1.26.4",
    "313": "2.0.1",
    "314": "2.3.2",
}
NUMPY_VERSION = NUMPY_VERSION_MAP[PYTHON_ABI_MAJ_MIN]
ORIGINAL_NUMPY_WHEEL = f"numpy-{NUMPY_VERSION}-{PYTHON_ABI}-linux_{PLATFORM}.whl"
SHOW_RE = re.compile(
    r'.*[\s](?P<wheel>\S+) is consistent with the following platform tag: "(?P<tag>\S+)".*',
    flags=re.DOTALL,
)
TAG_RE = re.compile(r"^manylinux_(?P<major>[0-9]+)_(?P<minor>[0-9]+)_(?P<arch>\S+)$")


class AnyLinuxContainer:
    def __init__(self, policy: str, tag: str, container: Container, io_folder: Path):
        self._policy = policy
        self._tag = tag
        self._container = container
        self._io_folder = io_folder

    @property
    def policy(self):
        return self._policy

    @property
    def io_folder(self):
        return self._io_folder

    def exec(
        self,
        cmd: str | list[str],
        *,
        expected_retcode: int = 0,
        cwd: str | None = None,
        env: dict[str, str] | None = None,
        coverage: bool = False,
    ) -> str:
        return docker_exec(
            self._container,
            cmd,
            expected_retcode=expected_retcode,
            cwd=cwd,
            env=env,
            coverage=coverage,
        )

    def show(
        self,
        wheel: str,
        *,
        verbose: bool = False,
        isa_ext_check: bool = True,
        expected_retcode: int = 0,
    ) -> str:
        isa_ext_check_arg = "" if isa_ext_check else "--disable-isa-ext-check"
        verbose_arg = "-v" if verbose else ""
        cmd = f"auditwheel {verbose_arg} show {isa_ext_check_arg} /io/{wheel}"
        return self.exec(cmd, expected_retcode=expected_retcode, coverage=True)

    def repair(
        self,
        wheel: str,
        *,
        isa_ext_check: bool = True,
        expected_retcode: int = 0,
        plat: str | None = None,
        only_plat: bool = True,
        strip: bool = False,
        library_paths: list[str] | None = None,
        excludes: list[str] | None = None,
    ) -> str:
        plat = plat or self._policy
        args = []
        if library_paths:
            ld_library_path = ":".join([*library_paths, "$LD_LIBRARY_PATH"])
            args.append(f"LD_LIBRARY_PATH={ld_library_path}")
        args.extend(["auditwheel", "repair", "-w", "/io", "--plat", plat])
        if only_plat:
            args.append("--only-plat")
        if not isa_ext_check:
            args.append("--disable-isa-ext-check")
        if strip:
            args.append("--strip")
        if excludes:
            args.extend(f"--exclude={exclude}" for exclude in excludes)
        args.append(f"/io/{wheel}")
        cmd = ["bash", "-c", " ".join(args)]
        return self.exec(cmd, expected_retcode=expected_retcode, coverage=True)

    def build_wheel(
        self,
        path: str,
        *,
        pre_wheel_cmd: str | None = None,
        env: dict[str, str] | None = None,
        check_filename: bool = True,
    ) -> str:
        args = ["if [ -d ./build ]; then rm -rf ./build ./*.egg-info; fi"]
        if pre_wheel_cmd:
            args.append(pre_wheel_cmd)
        args.append("python -m pip wheel --no-deps -w /io .")
        cmd = " && ".join(args)
        self.exec(["bash", "-c", cmd], cwd=path, env=env)

        filenames = os.listdir(self._io_folder)
        assert len(filenames) == 1
        orig_wheel = filenames[0]
        if check_filename:
            assert orig_wheel.endswith(f"-{PYTHON_ABI}-linux_{PLATFORM}.whl")
        assert "manylinux" not in orig_wheel
        assert "musllinux" not in orig_wheel
        return orig_wheel

    def check_wheel(
        self,
        name: str,
        *,
        version: str | None = None,
        python_abi: str | None = None,
        platform_tag: str | None = None,
    ) -> str:
        version = version or "0.0.1"
        python_abi = python_abi or PYTHON_ABI
        platform_tag = platform_tag or self._tag
        repaired_wheel = f"{name}-{version}-{python_abi}-{platform_tag}.whl"
        filenames = os.listdir(self._io_folder)
        assert len(filenames) == 2
        assert repaired_wheel in filenames
        return repaired_wheel

    @property
    def cache_dir(self) -> Path:
        return WHEEL_CACHE_FOLDER / self._policy


class PythonContainer:
    def __init__(self, container: Container):
        self._container = container

    def pip_install(self, args: str) -> str:
        cmd = f"pip install {args}"
        return self.exec(cmd)

    def install_wheel(self, filename: str) -> str:
        return self.pip_install(f"/io/{filename}")

    def run(self, cmd: str, *, env: dict[str, str] | None = None) -> str:
        args = ["python", "-c", cmd] if cmd.startswith(("from ", "import ")) else f"python {cmd}"
        return self.exec(args, env=env)

    def exec(
        self,
        cmd: str | list[str],
        expected_retcode: int = 0,
        env: dict[str, str] | None = None,
    ) -> str:
        return docker_exec(
            self._container,
            cmd,
            expected_retcode=expected_retcode,
            env=env,
        )


def find_src_folder() -> Path | None:
    candidate = HERE.parent.parent.resolve(strict=True)
    contents = os.listdir(candidate)
    if "pyproject.toml" in contents and "src" in contents:
        return candidate
    return None


def docker_start(
    image: str,
    volumes: dict[str, str] | None = None,
    env_variables: dict[str, str] | None = None,
) -> Container:
    """Start a long waiting idle program in container

    Return the container object to be used for 'docker exec' commands.
    """
    # Make sure to use the latest public version of the docker image
    if env_variables is None:
        env_variables = {}
    if volumes is None:
        volumes = {}
    client = docker.from_env()

    dvolumes = {host: {"bind": ctr, "mode": "rw"} for (ctr, host) in volumes.items()}
    goarch = {
        "x86_64": "amd64",
        "i686": "386",
        "aarch64": "arm64",
        "armv7l": "arm/v7",
    }.get(PLATFORM, PLATFORM)

    logger.info("Starting container with image %r", image)
    con = client.containers.run(
        image,
        ["tail", "-f", "/dev/null"],
        detach=True,
        volumes=dvolumes,
        environment=env_variables,
        platform=f"linux/{goarch}",
        working_dir="/auditwheel_src" if "/auditwheel_src" in volumes else None,
    )
    assert isinstance(con.id, str)
    logger.info("Started container %s", con.id[:12])
    return con


@contextmanager
def docker_container_ctx(
    image: str,
    io_dir: Path | None = None,
    env_variables: dict[str, str] | None = None,
) -> Generator[Container, None, None]:
    src_folder = find_src_folder()
    if src_folder is None:
        pytest.skip("Can only be run from the source folder")
    assert src_folder is not None

    if env_variables is None:
        env_variables = {}
    vols = {"/auditwheel_src": str(src_folder)}
    if io_dir is not None:
        vols["/io"] = str(io_dir)

    container = docker_start(image, vols, env_variables)
    try:
        yield container
    finally:
        container.remove(force=True)


def docker_exec(
    container: Container,
    cmd: str | list[str],
    *,
    expected_retcode: int = 0,
    cwd: str | None = None,
    env: dict[str, str] | None = None,
    coverage: bool = False,
) -> str:
    assert isinstance(container.id, str)
    logger.info("docker exec %s: %r", container.id[:12], cmd)
    # coverage is too slow with QEMU, even more so on ppc64le & s390x which lack a
    # C extension.
    # only enable coverage when QEMU is not enabled.
    if coverage and os.environ.get("AUDITWHEEL_QEMU", "") != "true":
        env = env.copy() if env is not None else {}
        env["COVERAGE_PROCESS_START"] = "/auditwheel_src/pyproject.toml"
    ec, raw_output = container.exec_run(cmd, workdir=cwd, environment=env)
    assert isinstance(raw_output, bytes)
    output = raw_output.decode("utf-8")
    assert isinstance(ec, int)
    if ec != expected_retcode:
        logger.info("docker exec error %s: %s", container.id[:12], output)
        raise CalledProcessError(ec, cmd, output=output)
    return output


@contextmanager
def tmp_docker_image(
    base: str,
    commands: list[str],
    setup_env: dict[str, str] | None = None,
) -> Generator[Any, None, None]:
    """Make a temporary docker image for tests

    Pulls the *base* image, runs *commands* inside it with *setup_env*, and
    commits the result as a new image. The image is removed on exiting the
    context.

    Making temporary images like this avoids each test having to re-run the
    same container setup steps.
    """
    if setup_env is None:
        setup_env = {}
    with docker_container_ctx(base, env_variables=setup_env) as con:
        for cmd in commands:
            docker_exec(con, cmd)
        image = con.commit()

    logger.info("Made image %s based on %s", image.short_id, base)
    try:
        yield image.id
    finally:
        client = image.client
        assert client is not None
        client.images.remove(image.id)


def assert_show_output(
    anylinux_ctr: AnyLinuxContainer,
    wheel: str,
    expected_tag: str,
    strict: bool,
    isa_ext_check: bool = True,
) -> None:
    output = anylinux_ctr.show(wheel, isa_ext_check=isa_ext_check)
    output = output.replace("\n", " ")
    match = SHOW_RE.match(output)
    assert match, f"{SHOW_RE.pattern!r} not found in:\n{output}"
    assert match["wheel"] == wheel
    if strict or "musllinux" in expected_tag:
        assert match["tag"] == expected_tag
    else:
        expected_match = TAG_RE.match(expected_tag)
        assert expected_match, f"No match for tag {expected_tag}"
        expected_glibc = (int(expected_match["major"]), int(expected_match["minor"]))
        actual_match = TAG_RE.match(match["tag"])
        assert actual_match, f"No match for tag {match['tag']}"
        actual_glibc = (int(actual_match["major"]), int(actual_match["minor"]))
        assert expected_match["arch"] == actual_match["arch"]
        assert actual_glibc <= expected_glibc


def build_numpy(container: AnyLinuxContainer, output_dir: Path) -> str:
    """Helper to build numpy from source using the specified container, into
    output_dir.
    """
    if container.policy.startswith("musllinux_"):
        container.exec("apk add openblas-dev")
        if container.policy.endswith("_s390x"):
            # https://github.com/numpy/numpy/issues/27932
            fix_hwcap = "echo '#define HWCAP_S390_VX 2048' >> /usr/include/bits/hwcap.h"
            container.exec(f'sh -c "{fix_hwcap}"')
    elif container.policy.startswith(("manylinux_2_12_", "manylinux_2_17_")):
        if tuple(int(part) for part in NUMPY_VERSION.split(".")[:2]) >= (1, 26):
            pytest.skip("numpy>=1.26 requires openblas")
        container.exec("yum install -y atlas atlas-devel")
    elif container.policy.startswith(("manylinux_2_31_", "manylinux_2_35_")):
        container.exec("apt-get install -y libopenblas-dev execstack")
        # TODO auditwheel shall check for executable stack:
        # https://github.com/pypa/auditwheel/issues/634
        container.exec(
            ["bash", "-c", "execstack -c $(find /usr/lib* -name 'libopenblas*.so')"],
        )
    else:
        container.exec("dnf install -y openblas-devel")

    cached_wheel = container.cache_dir / ORIGINAL_NUMPY_WHEEL
    orig_wheel = output_dir / ORIGINAL_NUMPY_WHEEL
    if cached_wheel.exists():
        # If numpy has already been built and put in cache, let's reuse this.
        shutil.copy2(cached_wheel, orig_wheel)
    else:
        # otherwise build the original linux_x86_64 numpy wheel from source
        # and put the result in the cache folder to speed-up future build.
        # This part of the build is independent of the auditwheel code-base
        # so it's safe to put it in cache.
        container.exec(f"pip wheel -w /io --no-binary=numpy numpy=={NUMPY_VERSION}")
        cached_wheel.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(orig_wheel, cached_wheel)
    return orig_wheel.name


class Anylinux:
    @pytest.fixture
    def io_folder(self, tmp_path):
        d = tmp_path / "io"
        d.mkdir(exist_ok=True)
        return d

    @pytest.fixture
    def python(
        self,
        docker_python_img: str,
        io_folder: Path,
    ) -> Generator[PythonContainer, None, None]:
        with docker_container_ctx(docker_python_img, io_folder) as container:
            yield PythonContainer(container)

    @pytest.fixture
    def anylinux(
        self,
        any_manylinux_img: tuple[str, str],
        io_folder: Path,
    ) -> Generator[AnyLinuxContainer, None, None]:
        policy, manylinux_img = any_manylinux_img
        env = {"PATH": PATH[policy]}
        with docker_container_ctx(manylinux_img, io_folder, env) as container:
            platform_tag = ".".join(
                sorted(
                    [f"{p}_{PLATFORM}" for p in [policy, *POLICY_ALIASES.get(policy, [])]],
                ),
            )
            yield AnyLinuxContainer(
                f"{policy}_{PLATFORM}",
                platform_tag,
                container,
                io_folder,
            )

    def test_numpy(self, anylinux: AnyLinuxContainer, python: PythonContainer) -> None:
        # Integration test: repair numpy built from scratch
        policy = anylinux.policy

        # First build numpy from source as a naive linux wheel that is tied
        # to system libraries (blas, libgfortran...)
        orig_wheel = build_numpy(anylinux, anylinux.io_folder)
        assert orig_wheel == ORIGINAL_NUMPY_WHEEL
        assert "manylinux" not in orig_wheel

        # Repair the wheel using the manylinux container
        anylinux.repair(orig_wheel)
        repaired_wheel = anylinux.check_wheel("numpy", version=NUMPY_VERSION)
        assert_show_output(anylinux, repaired_wheel, policy, False)

        # Check that the repaired numpy wheel can be installed and executed
        # on a modern linux image.
        python.install_wheel(repaired_wheel)
        output = python.run("/auditwheel_src/tests/integration/quick_check_numpy.py")
        assert output.strip() == "ok"

        # Check that numpy f2py works with a more recent version of gfortran
        if policy.startswith("musllinux_"):
            python.exec("apk add musl-dev gfortran")
        else:
            python.exec("apt-get install -y gfortran")
        if tuple(int(part) for part in NUMPY_VERSION.split(".")[:2]) >= (1, 26):
            python.pip_install("meson ninja")
            f2py_env = None
        else:
            f2py_env = {"SETUPTOOLS_USE_DISTUTILS": "stdlib"}
        python.run(
            "-m numpy.f2py -c /auditwheel_src/tests/integration/foo.f90 -m foo",
            env=f2py_env,
        )

        # Check that the 2 fortran runtimes are well isolated and can be loaded
        # at once in the same Python program:
        python.run("import numpy; import foo")

    def test_numpy_sbom(
        self,
        anylinux: AnyLinuxContainer,
        python: PythonContainer,
    ) -> None:
        policy = anylinux.policy
        if policy.startswith("manylinux_2_12_"):
            pytest.skip(f"whichprovides doesn't support {policy}")

        # Build and repair numpy
        orig_wheel = build_numpy(anylinux, anylinux.io_folder)
        assert orig_wheel == ORIGINAL_NUMPY_WHEEL
        assert "manylinux" not in orig_wheel

        # Repair the wheel using the manylinux container
        anylinux.repair(orig_wheel)
        repaired_wheel = anylinux.check_wheel("numpy", version=NUMPY_VERSION)
        assert_show_output(anylinux, repaired_wheel, policy, False)

        # Install the wheel so the SBOM document is added to the environment.
        python.install_wheel(repaired_wheel)

        # Load the auditwheel SBOM from .dist-info/sboms
        site_packages = python.run(
            "import site; print(site.getsitepackages()[0])",
        ).strip()
        assert re.match(
            r"\A/usr/local/lib/python[0-9.]+/site-packages\Z",
            site_packages,
        )
        sbom_data = python.run(
            f"-c \"print(open('{site_packages}/numpy-{NUMPY_VERSION}.dist-info/sboms/auditwheel.cdx.json').read())\"",  # noqa: E501
        ).strip()
        sbom = json.loads(sbom_data)

        # Separate all the components that vary over test runs.
        sbom_tools = sbom["metadata"].pop("tools")
        sbom_components = sbom.pop("components")
        sbom_dependencies = sbom.pop("dependencies")

        expected_numpy_purl = f"pkg:pypi/numpy@{NUMPY_VERSION}?file_name={repaired_wheel}"
        assert sbom == {
            "bomFormat": "CycloneDX",
            "specVersion": "1.4",
            "version": 1,
            "metadata": {
                "component": {
                    "type": "library",
                    "bom-ref": expected_numpy_purl,
                    "name": "numpy",
                    "version": NUMPY_VERSION,
                    "purl": expected_numpy_purl,
                },
                # "tools": [{...}, ...],
            },
            # "components": [{...}, ...],
            # "dependencies": [{...}, ...],
        }

        assert len(sbom_tools) == 1
        assert sbom_tools[0]["name"] == "auditwheel"
        assert "version" in sbom_tools[0]

        assert sbom_components[0] == {
            "bom-ref": expected_numpy_purl,
            "name": "numpy",
            "purl": expected_numpy_purl,
            "type": "library",
            "version": NUMPY_VERSION,
        }

        component_purls = {component["purl"].split("@")[0] for component in sbom_components[1:]}
        # Package URL prefixes must match for a policy.
        if policy.startswith("musllinux"):
            expected_purl_prefix = "pkg:apk/alpine/"
        elif policy.startswith("manylinux_2_17_"):
            expected_purl_prefix = "pkg:rpm/centos/"
        elif policy.startswith(("manylinux_2_31_", "manylinux_2_35_")):
            expected_purl_prefix = "pkg:deb/ubuntu/"
        elif policy == "manylinux_2_39_riscv64":
            expected_purl_prefix = "pkg:rpm/rocky/"
        else:
            expected_purl_prefix = "pkg:rpm/almalinux/"

        assert all(purl.startswith(expected_purl_prefix) for purl in component_purls), str(
            component_purls,
        )

        # We expect libgfortran and openblas* (aka atlas) as dependencies.
        if policy != "musllinux_1_2_riscv64":
            assert any("libgfortran" in purl for purl in component_purls), str(
                component_purls,
            )
        assert any("openblas" in purl for purl in component_purls) or any(
            "atlas" in purl for purl in component_purls
        ), str(component_purls)

        assert len(sbom_dependencies) == len(component_purls) + 1

    def test_exclude(self, anylinux: AnyLinuxContainer) -> None:
        """Test the --exclude argument to avoid grafting certain libraries."""
        test_path = "/auditwheel_src/tests/integration/testrpath"
        orig_wheel = anylinux.build_wheel(test_path)

        output = anylinux.repair(
            orig_wheel,
            excludes=["liba.so"],
            library_paths=[f"{test_path}/a"],
        )
        assert "Excluding liba.so" in output
        repaired_wheel = anylinux.check_wheel("testrpath")

        # Make sure we don't have liba.so & libb.so in the result
        contents = zipfile.ZipFile(anylinux.io_folder / repaired_wheel).namelist()
        assert not any(x for x in contents if "/liba" in x or "/libb" in x)

    def test_with_binary_executable(
        self,
        anylinux: AnyLinuxContainer,
        python: PythonContainer,
    ) -> None:
        # Test building a wheel that contains a binary executable (e.g., a program)
        policy = anylinux.policy
        if policy.startswith("musllinux_"):
            anylinux.exec("apk add gsl-dev")
        elif policy.startswith(("manylinux_2_31_", "manylinux_2_35_")):
            anylinux.exec("apt-get install -y libgsl-dev")
        elif policy == "manylinux_2_39_riscv64":
            pytest.skip(reason=f"no gsl-devel on {policy}")
        else:
            anylinux.exec("yum install -y gsl-devel")

        test_path = "/auditwheel_src/tests/integration/testpackage"
        orig_wheel = anylinux.build_wheel(test_path, check_filename=False)
        assert orig_wheel == "testpackage-0.0.1-py3-none-any.whl"

        # manylinux_2_34_x86_64 uses x86_64_v2 for this test
        isa_ext_check = policy != "manylinux_2_34_x86_64"

        # Repair the wheel using the appropriate manylinux container
        anylinux.repair(orig_wheel, isa_ext_check=isa_ext_check)
        repaired_wheel = anylinux.check_wheel("testpackage", python_abi="py3-none")
        assert_show_output(anylinux, repaired_wheel, policy, False, isa_ext_check)

        python.install_wheel(repaired_wheel)
        output = python.run("from testpackage import runit; print(runit(1.5))")
        assert output.strip() == "2.25"

        # Both testprogram and testprogram_nodeps square a number, but:
        #   * testprogram links against libgsl and had to have its RPATH
        #     rewritten.
        #   * testprogram_nodeps links against no shared libraries and wasn't
        #     rewritten.
        #
        # Both executables should work when called from the installed bin directory.
        assert python.exec(["/usr/local/bin/testprogram", "4"]) == "16\n"
        assert python.exec(["/usr/local/bin/testprogram_nodeps", "4"]) == "16\n"

        # testprogram should be a Python shim since we had to rewrite its RPATH.
        shebang = python.exec(["head", "-n1", "/usr/local/bin/testprogram"])
        assert shebang in {
            "#!/usr/local/bin/python\n",
            "#!/usr/local/bin/python3\n",
            f"#!/usr/local/bin/python{'.'.join(PYTHON_MAJ_MIN)}\n",
        }

        # testprogram_nodeps should be the unmodified ELF binary.
        assert python.exec(["head", "-c4", "/usr/local/bin/testprogram_nodeps"]) == "\x7fELF"

    def test_pure_wheel(self, anylinux: AnyLinuxContainer) -> None:
        anylinux.exec(
            "pip download --no-deps -d /io --only-binary=:all: six==1.16.0",
        )
        orig_wheel = "six-1.16.0-py2.py3-none-any.whl"
        assert anylinux.io_folder.joinpath(orig_wheel).exists()

        # Repair the wheel using the manylinux container
        output = anylinux.repair(orig_wheel, expected_retcode=1)
        assert "This does not look like a platform wheel" in output

        output = anylinux.show(orig_wheel, expected_retcode=1)
        assert "This does not look like a platform wheel" in output

    @pytest.mark.parametrize("dtag", ["none", "rpath", "runpath"])
    def test_rpath(
        self,
        anylinux: AnyLinuxContainer,
        python: PythonContainer,
        dtag: str,
    ) -> None:
        # Test building a wheel that contains an extension depending on a library
        # with RPATH or RUNPATH set.
        # Following checks are performed:
        # - check if RUNPATH is replaced by RPATH if the library has other grafted deps
        # - check if RPATH location is correct, i.e. it is inside .libs directory
        #   where all gathered libraries are put
        # - check a library with no deps has no RUNPATH/RPATH tags

        policy = anylinux.policy

        test_path = "/auditwheel_src/tests/integration/testrpath"
        orig_wheel = anylinux.build_wheel(test_path, env={"DTAG": dtag})

        for lib in ["a", "b"]:
            with HERE.joinpath("testrpath", lib, f"lib{lib}.so").open("rb") as f:
                elf = ELFFile(f)
                dynamic = elf.get_section_by_name(".dynamic")
                tags = {t.entry.d_tag for t in dynamic.iter_tags()}
                if dtag != "none":
                    assert f"DT_{dtag.upper()}" in tags
                else:
                    assert "DT_RPATH" not in tags
                    assert "DT_RUNPATH" not in tags

        # Repair the wheel using the appropriate manylinux container
        library_paths = [f"{test_path}/a"]
        if dtag == "none":
            library_paths.append(f"{test_path}/b")
        anylinux.repair(orig_wheel, library_paths=library_paths)
        repaired_wheel = anylinux.check_wheel("testrpath")
        assert_show_output(anylinux, repaired_wheel, policy, False)

        python.install_wheel(repaired_wheel)
        output = python.run("from testrpath import testrpath; print(testrpath.func())")
        assert output.strip() == "11"
        with zipfile.ZipFile(anylinux.io_folder / repaired_wheel) as w:
            libraries = tuple(name for name in w.namelist() if "testrpath.libs/lib" in name)
            assert len(libraries) == 2
            assert any(".libs/liba" in name for name in libraries)
            assert any(".libs/libb" in name for name in libraries)
            for name in libraries:
                with w.open(name) as f:
                    elf = ELFFile(io.BytesIO(f.read()))
                    dynamic = elf.get_section_by_name(".dynamic")
                    # DT_RUNPATH shall be removed
                    runpath_tags = [t for t in dynamic.iter_tags() if t.entry.d_tag == "DT_RUNPATH"]
                    assert len(runpath_tags) == 0
                    rpath_tags = [t for t in dynamic.iter_tags() if t.entry.d_tag == "DT_RPATH"]
                    if ".libs/liba" in name:
                        # liba has a dependency on libb
                        # DT_RPATH shall be overridden or written with "$ORIGIN"
                        assert len(rpath_tags) == 1
                        assert rpath_tags[0].rpath == "$ORIGIN"
                    if ".libs/libb" in name:
                        # libb has no dependency
                        # DT_RPATH shall be removed
                        assert len(rpath_tags) == 0

    def test_partialresolution(self, anylinux: AnyLinuxContainer, python: PythonContainer) -> None:

        policy = anylinux.policy

        test_path = "/auditwheel_src/tests/integration/testpartialresolution"
        orig_wheel = anylinux.build_wheel(test_path)

        # Repair the wheel using the appropriate manylinux container
        anylinux.repair(orig_wheel)
        repaired_wheel = anylinux.check_wheel("testpartialresolution")
        assert_show_output(anylinux, repaired_wheel, policy, False)

        python.install_wheel(repaired_wheel)
        output = python.run(
            "from testpartialresolution import testpartialresolution;"
            "print(testpartialresolution.func())",
        )
        assert output.strip() == "11"

        with zipfile.ZipFile(anylinux.io_folder / repaired_wheel) as w:
            libraries = tuple(name for name in w.namelist() if "testpartialresolution/lib" in name)
            assert len(libraries) == 2
            for name in libraries:
                with w.open(name) as f:
                    elf = ELFFile(io.BytesIO(f.read()))
                    dynamic = elf.get_section_by_name(".dynamic")
                    runpath_tags = [t for t in dynamic.iter_tags() if t.entry.d_tag == "DT_RUNPATH"]
                    assert len(runpath_tags) == 0
                    rpath_tags = [t for t in dynamic.iter_tags() if t.entry.d_tag == "DT_RPATH"]
                    assert len(rpath_tags) == 0

    def test_multiple_top_level(
        self,
        anylinux: AnyLinuxContainer,
        python: PythonContainer,
    ) -> None:
        policy = anylinux.policy

        test_path = "/auditwheel_src/tests/integration/multiple_top_level"
        orig_wheel = anylinux.build_wheel(test_path, pre_wheel_cmd="make clean all")

        # Repair the wheel using the appropriate manylinux container
        anylinux.repair(
            orig_wheel,
            library_paths=[f"{test_path}/lib-src/a", f"{test_path}/lib-src/b"],
        )
        repaired_wheel = anylinux.check_wheel("multiple_top_level")
        assert_show_output(anylinux, repaired_wheel, policy, False)

        python.install_wheel(repaired_wheel)
        for mod, func, expected in [
            ("example_a", "example_a", "11"),
            ("example_b", "example_b", "110"),
        ]:
            output = python.run(f"from {mod} import {func}; print({func}())").strip()
            assert output.strip() == expected
        with zipfile.ZipFile(anylinux.io_folder / repaired_wheel) as w:
            for lib_name in ["liba", "libb"]:
                assert any(
                    re.match(rf"multiple_top_level.libs/{lib_name}.*\.so", name)
                    for name in w.namelist()
                )

    def test_internal_rpath(
        self,
        anylinux: AnyLinuxContainer,
        python: PythonContainer,
    ) -> None:
        policy = anylinux.policy

        test_path = "/auditwheel_src/tests/integration/internal_rpath"
        orig_wheel = anylinux.build_wheel(
            test_path,
            pre_wheel_cmd="make clean all && mv lib-src/a/liba.so internal_rpath",
        )

        # Repair the wheel using the appropriate manylinux container
        anylinux.repair(orig_wheel, library_paths=[f"{test_path}/lib-src/b"])
        repaired_wheel = anylinux.check_wheel("internal_rpath")
        assert_show_output(anylinux, repaired_wheel, policy, False)

        python.install_wheel(repaired_wheel)
        for mod, func, expected in [
            ("example_a", "example_a", "11"),
            ("example_b", "example_b", "10"),
        ]:
            output = python.run(
                f"from internal_rpath.{mod} import {func}; print({func}())",
            )
            assert output.strip() == expected
        with zipfile.ZipFile(anylinux.io_folder / repaired_wheel) as w:
            for lib_name in ["libb"]:
                assert any(
                    re.match(rf"internal_rpath.libs/{lib_name}.*\.so", name)
                    for name in w.namelist()
                )

    def test_strip(self, anylinux: AnyLinuxContainer, python: PythonContainer) -> None:
        policy = anylinux.policy

        test_path = "/auditwheel_src/tests/integration/sample_extension"
        orig_wheel = anylinux.build_wheel(test_path)
        assert orig_wheel.startswith("sample_extension-0.1.0")

        # Repair the wheel using the appropriate manylinux container
        anylinux.repair(orig_wheel, strip=True)

        repaired_wheel = next(anylinux.io_folder.glob(f"*{policy}*.whl")).name

        python.install_wheel(repaired_wheel)
        output = python.run(
            "from sample_extension import test_func; print(test_func(1))",
        )
        assert output.strip() == "2"

    def test_nonpy_rpath(
        self,
        anylinux: AnyLinuxContainer,
        python: PythonContainer,
    ) -> None:
        # Tests https://github.com/pypa/auditwheel/issues/136
        policy = anylinux.policy

        test_path = "/auditwheel_src/tests/integration/nonpy_rpath"
        orig_wheel = anylinux.build_wheel(test_path)

        # Repair the wheel using the appropriate manylinux container
        anylinux.repair(orig_wheel)
        repaired_wheel = anylinux.check_wheel("nonpy_rpath")
        assert_show_output(anylinux, repaired_wheel, policy, False)

        # Test the resulting wheel outside the manylinux container
        python.install_wheel(repaired_wheel)
        python.run(
            "import nonpy_rpath; assert nonpy_rpath.crypt_something().startswith('*')",
        )

    def test_glibcxx_3_4_25(
        self,
        anylinux: AnyLinuxContainer,
        python: PythonContainer,
    ) -> None:
        policy = anylinux.policy

        test_path = "/auditwheel_src/tests/integration/test_glibcxx_3_4_25"
        orig_wheel = anylinux.build_wheel(test_path)
        assert orig_wheel.startswith("testentropy-0.0.1")

        # Repair the wheel using the appropriate manylinux container
        if policy.startswith("manylinux_2_28_"):
            with pytest.raises(CalledProcessError):
                anylinux.repair(orig_wheel)
            # TODO if a "permissive" mode is implemented, add the relevant flag to the
            # repair_command here and drop the return statement below
            return

        anylinux.repair(orig_wheel)

        repaired_wheel = next(anylinux.io_folder.glob(f"*{policy}*.whl")).name

        python.install_wheel(repaired_wheel)
        python.run("from testentropy import run; exit(run())")

    @pytest.mark.skipif(
        PLATFORM != "x86_64",
        reason="ISA extension only implemented on x86_64",
    )
    @pytest.mark.parametrize("isa_ext", ["x86-64-v2", "x86-64-v3", "x86-64-v4"])
    def test_isa_variants(self, anylinux: AnyLinuxContainer, isa_ext: str) -> None:
        policy = anylinux.policy
        if policy.startswith(("manylinux_2_12_", "manylinux_2_17_")):
            pytest.skip("skip old gcc")

        test_path = "/auditwheel_src/tests/integration/testdependencies"
        orig_wheel = anylinux.build_wheel(
            test_path,
            env={"WITH_DEPENDENCY": "1", "WITH_ARCH": isa_ext},
        )

        # repair failure with ISA check
        with pytest.raises(CalledProcessError):
            anylinux.repair(orig_wheel, library_paths=[test_path])

        anylinux.repair(orig_wheel, library_paths=[test_path], isa_ext_check=False)
        repaired_wheel = anylinux.check_wheel("testdependencies", platform_tag=policy)
        assert_show_output(anylinux, repaired_wheel, policy, True, False)

        # with ISA check, we shall not report a manylinux/musllinux policy
        assert_show_output(anylinux, repaired_wheel, f"linux_{PLATFORM}", True)

    @pytest.mark.parametrize(
        "arch",
        [
            Architecture.aarch64,
            Architecture.armv7l,
            Architecture.i686,
            Architecture.ppc64le,
            Architecture.riscv64,
            Architecture.s390x,
            Architecture.x86_64,
        ],
    )
    @pytest.mark.parametrize("libc", [Libc.GLIBC, Libc.MUSL])
    def test_cross_repair(
        self,
        anylinux: AnyLinuxContainer,
        libc: Libc,
        arch: Architecture,
    ) -> None:
        if libc == Libc.MUSL:
            source = "musllinux_1_2"
            platform_tag = f"musllinux_1_2_{arch.value}"
            python_abi = "cp312-cp312"
        else:
            assert libc == Libc.GLIBC
            source = "glibc"
            platform_tag = f"manylinux2014_{arch.value}.manylinux_2_17_{arch.value}"
            if arch in {Architecture.x86_64, Architecture.i686}:
                platform_tag = f"manylinux1_{arch.value}.manylinux_2_5_{arch.value}"
            elif arch == Architecture.riscv64:
                platform_tag = f"manylinux_2_31_{arch.value}"
            python_abi = "cp313-cp313"
        test_path = f"/auditwheel_src/tests/integration/arch-wheels/{source}"
        orig_wheel = f"testsimple-0.0.1-{python_abi}-linux_{arch.value}.whl"
        anylinux.exec(["cp", "-f", f"{test_path}/{orig_wheel}", f"/io/{orig_wheel}"])
        anylinux.repair(orig_wheel, plat="auto", only_plat=False)
        anylinux.check_wheel(
            "testsimple",
            python_abi=python_abi,
            platform_tag=platform_tag,
        )


class TestManylinux(Anylinux):
    @pytest.fixture(scope="session")
    def docker_python_img(self):
        """The glibc Python base image with up-to-date pip"""
        commnds = ["pip install -U pip", "apt-get update -yqq"]
        with tmp_docker_image(MANYLINUX_PYTHON_IMAGE_ID, commnds) as img_id:
            yield img_id

    @pytest.fixture(scope="session", params=MANYLINUX_IMAGES.keys())
    def any_manylinux_img(self, request):
        """Each manylinux image, with auditwheel installed.

        Plus up-to-date pip, setuptools and coverage
        """
        policy = request.param
        check_set = {
            "manylinux_2_12": {"38", "39", "310"},
        }.get(policy)
        if check_set and PYTHON_ABI_MAJ_MIN not in check_set:
            pytest.skip(f"{policy} images do not support cp{PYTHON_ABI_MAJ_MIN}")

        base = MANYLINUX_IMAGES[policy]
        env = {"PATH": PATH[policy]}
        commands = [
            'git config --global --add safe.directory "/auditwheel_src"',
            "pip install -U pip setuptools 'coverage[toml]>=7.13'",
            "pip install -U -e /auditwheel_src",
        ]
        if policy in {"manylinux_2_31", "manylinux_2_35"}:
            commands.append("apt-get update -yqq")
        with tmp_docker_image(base, commands, env) as img_id:
            yield policy, img_id

    @pytest.mark.parametrize("with_dependency", ["0", "1"])
    def test_image_dependencies(
        self,
        with_dependency: str,
        anylinux: AnyLinuxContainer,
        python: PythonContainer,
    ) -> None:
        # try to repair the wheel targeting different policies
        #
        # with_dependency == 0
        #   The python module has no dependencies that should be grafted-in and
        #   uses versioned symbols not available on policies pre-dating the policy
        #   matching the image being tested.
        # with_dependency == 1
        #   The python module itself does not use versioned symbols but has a
        #   dependency that should be grafted-in that uses versioned symbols not
        #   available on policies pre-dating the policy matching the image being
        #   tested.

        policy_name = anylinux.policy

        test_path = "/auditwheel_src/tests/integration/testdependencies"
        orig_wheel = anylinux.build_wheel(
            test_path,
            env={"WITH_DEPENDENCY": with_dependency},
        )

        policies = WheelPolicies(libc=Libc.GLIBC, arch=Architecture(PLATFORM))
        policy = policies.get_policy_by_name(policy_name)
        older_policies = [
            f"{p}_{PLATFORM}"
            for p in MANYLINUX_IMAGES
            if policy < policies.get_policy_by_name(f"{p}_{PLATFORM}")
        ]
        for target_policy in older_policies:
            # we shall fail to repair the wheel when targeting an older policy than
            # the one matching the image
            with pytest.raises(CalledProcessError):
                anylinux.repair(
                    orig_wheel,
                    plat=target_policy,
                    library_paths=[test_path],
                )

        # check all works properly when targeting the policy matching the image
        # use "auto" platform
        anylinux.repair(
            orig_wheel,
            only_plat=False,
            plat="auto",
            library_paths=[test_path],
        )
        repaired_wheel = anylinux.check_wheel("testdependencies")
        # we shall only get the current policy tag with "auto" platform
        assert_show_output(anylinux, repaired_wheel, policy_name, True)

        # check the original wheel with a dependency was not compliant
        # and check the one without a dependency was already compliant
        expected = f"linux_{PLATFORM}" if with_dependency == "1" else policy_name
        assert_show_output(anylinux, orig_wheel, expected, True)

        python.install_wheel(repaired_wheel)
        python.run("from testdependencies import run; exit(run())")

    @pytest.mark.parametrize(
        "target_policy",
        [f"{p}_{PLATFORM}" for p in list(MANYLINUX_IMAGES.keys())]
        + [f"{p}_{PLATFORM}" for aliases in POLICY_ALIASES.values() for p in aliases],
    )
    @pytest.mark.parametrize("only_plat", [True, False])
    def test_compat(
        self,
        target_policy: str,
        only_plat: bool,
        anylinux: AnyLinuxContainer,
        python: PythonContainer,
    ) -> None:
        # test building wheels with compatibility with older spec
        # check aliases for older spec
        test_path = "/auditwheel_src/tests/integration/testsimple"
        orig_wheel = anylinux.build_wheel(test_path)

        if PLATFORM in {"x86_64", "i686"}:
            expect = f"manylinux_2_5_{PLATFORM}"
            expect_tag = f"manylinux1_{PLATFORM}.manylinux_2_5_{PLATFORM}"
        elif PLATFORM == "riscv64":
            expect = f"manylinux_2_31_{PLATFORM}"
            expect_tag = f"manylinux_2_31_{PLATFORM}"
        else:
            expect = f"manylinux_2_17_{PLATFORM}"
            expect_tag = f"manylinux2014_{PLATFORM}.manylinux_2_17_{PLATFORM}"

        target_tag = target_policy
        for pep600_policy, aliases in POLICY_ALIASES.items():
            policy_ = f"{pep600_policy}_{PLATFORM}"
            aliases_ = [f"{p}_{PLATFORM}" for p in aliases]
            if target_policy == policy_ or target_policy in aliases_:
                target_tag = ".".join(sorted([policy_, *aliases_]))

        # we shall ba able to repair the wheel for all targets
        anylinux.repair(orig_wheel, plat=target_policy, only_plat=only_plat)
        if only_plat or target_tag == expect_tag:
            repaired_tag = target_tag
        else:
            repaired_tag = ".".join(
                sorted(expect_tag.split(".") + target_tag.split(".")),
            )
        repaired_wheel = anylinux.check_wheel("testsimple", platform_tag=repaired_tag)

        assert_show_output(anylinux, repaired_wheel, expect, True)

        with zipfile.ZipFile(anylinux.io_folder / repaired_wheel) as z:
            for file in z.namelist():
                assert not file.startswith("testsimple.libs"), "should not have empty .libs folder"

        python.install_wheel(repaired_wheel)
        python.run("from testsimple import run; exit(run())")

    @pytest.mark.skipif(
        PLATFORM != "x86_64",
        reason=f"libmvec not supported on {PLATFORM}",
    )
    def test_mvec(self, anylinux: AnyLinuxContainer, python: PythonContainer) -> None:
        # Tests https://github.com/pypa/auditwheel/issues/645
        policy = anylinux.policy

        if policy.startswith(("manylinux_2_12_", "manylinux_2_17_")):
            pytest.skip(f"libmvec not supported on {policy}")

        test_path = "/auditwheel_src/tests/integration/test_mvec"
        orig_wheel = anylinux.build_wheel(test_path, check_filename=False)

        # Repair the wheel using the appropriate manylinux container
        anylinux.repair(orig_wheel, only_plat=False)
        platform_tag = f"manylinux_2_24_x86_64.{policy}"
        repaired_wheel = anylinux.check_wheel(
            "test_mvec",
            python_abi="py3-none",
            platform_tag=platform_tag,
        )
        assert_show_output(anylinux, repaired_wheel, policy, False)

        # Test the resulting wheel outside the manylinux container
        python.install_wheel(repaired_wheel)

    def test_zlib_blacklist(self, anylinux: AnyLinuxContainer) -> None:
        policy = anylinux.policy
        if policy.startswith(
            (
                "manylinux_2_17_",
                "manylinux_2_28_",
                "manylinux_2_31_",
                "manylinux_2_34_",
                "manylinux_2_35_",
                "manylinux_2_39_",
            ),
        ):
            pytest.skip(f"{policy} image has no blacklist symbols in libz.so.1")

        test_path = "/auditwheel_src/tests/integration/testzlib"
        orig_wheel = anylinux.build_wheel(test_path)
        assert orig_wheel.startswith("testzlib-0.0.1")

        # Repair the wheel using the appropriate manylinux container
        with pytest.raises(CalledProcessError):
            anylinux.repair(orig_wheel)

        # Check auditwheel show warns about the black listed symbols
        output = anylinux.show(orig_wheel, verbose=True)
        assert "black-listed symbol dependencies" in output.replace("\n", " ")

    @pytest.mark.skipif(PLATFORM != "aarch64", reason="glibc blacklist only for aarch64")
    def test_glibc_blacklist(self, anylinux: AnyLinuxContainer, python: PythonContainer) -> None:
        # https://github.com/pypa/auditwheel/issues/647
        policy = anylinux.policy
        if policy.startswith("manylinux_2_17_"):
            pytest.skip(f"{policy} image can't build this test")

        test_path = "/auditwheel_src/tests/integration/test_aarch64_glibc_blacklist"
        orig_wheel = anylinux.build_wheel(test_path, check_filename=False)
        assert orig_wheel.startswith("test_aarch64_glibc_blacklist-0.0.1-")

        # Repair the wheel using the appropriate manylinux container
        anylinux.repair(orig_wheel, only_plat=not policy.startswith("manylinux_2_28_"))
        platform_tag = policy
        if policy.startswith("manylinux_2_28_"):
            platform_tag = f"manylinux_2_24_aarch64.{platform_tag}"
        repaired_wheel = anylinux.check_wheel(
            "test_aarch64_glibc_blacklist",
            python_abi="py3-none",
            platform_tag=platform_tag,
        )

        # Check auditwheel show warns about the black listed symbols
        output = anylinux.show(orig_wheel, verbose=True)
        output = output.replace("\n", " ")
        assert "black-listed symbol dependencies" in output
        # glibc moves symbol from library to library so libc.so.6 is used as a catch all
        # as it's more likely to be needed
        needle = (
            "From libc.so.6: __cxa_thread_atexit_impl, __issignaling, __issignalingf, "
            "__issignalingl, pthread_getattr_default_np, pthread_setattr_default_np"
        )
        assert needle in output
        needle = "From libm.so.6: __issignaling, __issignalingf, __issignalingl"
        assert needle in output
        needle = "From libpthread.so.0: pthread_getattr_default_np, pthread_setattr_default_np"
        if policy.startswith("manylinux_2_28_"):
            assert needle in output
        else:
            assert needle not in output

        # Test the resulting wheel outside the manylinux container
        python.install_wheel(repaired_wheel)
        python.exec("test_signaling")
        python.exec("test_signalingf")
        python.exec("test_signalingl")
        python.exec("test_pthread")


class TestMusllinux(Anylinux):
    @pytest.fixture(scope="session")
    def docker_python_img(self):
        """The alpine Python base image with up-to-date pip"""
        commands = ["pip install -U pip"]
        with tmp_docker_image(MUSLLINUX_PYTHON_IMAGE_ID, commands) as img_id:
            yield img_id

    @pytest.fixture(scope="session", params=MUSLLINUX_IMAGES.keys())
    def any_manylinux_img(self, request):
        """Each musllinux image, with auditwheel installed.

        Plus up-to-date pip, setuptools and coverage
        """
        policy = request.param
        base = MUSLLINUX_IMAGES[policy]
        env = {"PATH": PATH[policy]}
        commands = [
            'git config --global --add safe.directory "/auditwheel_src"',
            "pip install -U pip setuptools 'coverage[toml]>=7.13'",
            "pip install -U -e /auditwheel_src",
        ]
        with tmp_docker_image(base, commands, env) as img_id:
            yield policy, img_id



from __future__ import annotations

import json
import logging
import re
from collections import defaultdict
from dataclasses import dataclass, replace
from pathlib import Path
from typing import TYPE_CHECKING, Any

from auditwheel.architecture import Architecture
from auditwheel.elfutils import filter_undefined_symbols
from auditwheel.error import InvalidLibcError
from auditwheel.lddtree import LIBPYTHON_RE
from auditwheel.libc import Libc
from auditwheel.tools import is_subdir
from auditwheel.wheeltools import android_api_level, get_wheel_platforms

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable

    from auditwheel.lddtree import DynamicExecutable

_HERE = Path(__file__).parent
_MUSL_POLICY_RE = re.compile(r"^musllinux_\d+_\d+$")

logger = logging.getLogger(__name__)

_POLICY_JSON_MAP = {
    Libc.GLIBC: _HERE / "manylinux-policy.json",
    Libc.MUSL: _HERE / "musllinux-policy.json",
    Libc.ANDROID: _HERE / "android-policy.json",
}


@dataclass(frozen=True)
class Policy:
    name: str
    aliases: tuple[str, ...]
    priority: int
    symbol_versions: dict[str, frozenset[str]]
    whitelist: frozenset[str]
    blacklist: dict[str, frozenset[str]]

    def __lt__(self, other: Any) -> bool:  # noqa: ANN401
        if not isinstance(other, Policy):
            raise NotImplementedError
        return self.priority < other.priority


@dataclass()
class ExternalReference:
    libs: dict[str, Path | None]
    blacklist: dict[str, list[str]]
    policy: Policy


class WheelPolicies:
    def __init__(
        self,
        *,
        libc: Libc,
        arch: Architecture,
        wheel_fn: Path | None = None,
        musl_policy: str | None = None,
    ) -> None:
        if libc != Libc.MUSL and musl_policy is not None:
            msg = f"'musl_policy' shall be None for libc {libc.name}"
            raise ValueError(msg)
        if libc == Libc.MUSL:
            if musl_policy is None:
                try:
                    musl_version = libc.get_current_version()
                    musl_policy = f"musllinux_{musl_version.major}_{musl_version.minor}"
                except InvalidLibcError:
                    logger.warning(
                        "can't determine musl libc version, latest known version will be used.",
                    )
            elif _MUSL_POLICY_RE.match(musl_policy) is None:
                msg = f"Invalid 'musl_policy': '{musl_policy}'"
                raise ValueError(msg)
        policies = json.loads(_POLICY_JSON_MAP[libc].read_text())
        self._policies: list[Policy] = []
        self._architecture = arch
        self._libc_variant = libc
        self._musl_policy = musl_policy

        base_arch = arch.baseline.value
        _validate_pep600_compliance(policies)
        for policy in policies:
            if self._musl_policy is not None and policy["name"] not in {
                "linux",
                self._musl_policy,
            }:
                continue
            if arch.value in policy["symbol_versions"] or policy["name"] == "linux":
                name = f"{policy['name']}_{base_arch}"
                if policy["name"] != "linux":
                    symbol_versions = {
                        k: frozenset(v) for k, v in policy["symbol_versions"][base_arch].items()
                    }
                else:
                    symbol_versions = {}
                aliases = tuple(f"{alias}_{base_arch}" for alias in policy["aliases"])
                whitelist = _fixup_musl_libc_soname(libc, arch, policy["lib_whitelist"])
                blacklist = {k: frozenset(v) for k, v in policy["blacklist"].items()}
                policy_ = Policy(
                    name=name,
                    aliases=aliases,
                    priority=policy["priority"],
                    symbol_versions=symbol_versions,
                    whitelist=whitelist,
                    blacklist=blacklist,
                )
                self._policies.append(policy_)

        self._policies.sort()

        if self._libc_variant == Libc.MUSL:
            if musl_policy is None:
                self._musl_policy = "_".join(self._policies[1].name.split("_")[0:3])
                self._policies = [self._policies[0], self._policies[1]]
            assert len(self._policies) == 2, self._policies  # noqa: S101

        elif self._libc_variant == Libc.ANDROID:
            # Android wheels always have a platform tag with an API level (OS version). We won't
            # change that tag, we'll only use it to determine which libraries need to be grafted.
            if wheel_fn is None:
                msg = "wheel_fn is required when selecting Android policies"
                raise ValueError(msg)

            platforms = get_wheel_platforms(wheel_fn.name)
            if len(platforms) != 1:
                msg = f"Android wheels must have exactly one platform tag, got {platforms}"
                raise ValueError(msg)
            platform = platforms[0]
            api_level = android_api_level(platform)

            # One or two new API levels are created every year, so the policy file only lists the
            # levels in which the available libraries changed
            # (https://developer.android.com/ndk/guides/stable_apis). We should use the policy with
            # the highest API level that's less than or equal to the wheel's tag.
            for p in self._policies:
                if p.name.startswith("android") and android_api_level(p.name) <= api_level:
                    # Rename the policy to match the existing tag, and remove all other policies.
                    self._policies = [self.linux, replace(p, name=platform)]
                    break
            else:
                # The minimum API level is 21, which was the first version to be officially
                # supported by Python (see PEP 738).
                msg = f"no Android policies match the platform tag {platform!r}"
                raise ValueError(msg)

    def __iter__(self) -> Generator[Policy]:
        yield from self._policies

    @property
    def libc(self) -> Libc:
        return self._libc_variant

    @property
    def architecture(self) -> Architecture:
        return self._architecture

    @property
    def highest(self) -> Policy:
        return self._policies[-1]

    @property
    def lowest(self) -> Policy:
        """Lowest policy that's not linux"""
        return self._policies[1]

    @property
    def linux(self) -> Policy:
        return self._policies[0]

    def get_policy_by_name(self, name: str) -> Policy:
        matches = [p for p in self._policies if p.name == name or name in p.aliases]
        if len(matches) == 0:
            msg = f"no policy named {name!r} found"
            raise LookupError(msg)
        if len(matches) > 1:
            msg = "Internal error. Policies should be unique"
            raise RuntimeError(msg)
        return matches[0]

    def versioned_symbols_policy(
        self,
        versioned_symbols: dict[str, set[str]],
    ) -> Policy:
        def policy_is_satisfied(
            policy_name: str,
            policy_sym_vers: dict[str, set[str]],
        ) -> bool:
            policy_satisfied = True
            for name in set(required_vers) & set(policy_sym_vers):
                if not required_vers[name].issubset(policy_sym_vers[name]):
                    for symbol in required_vers[name] - policy_sym_vers[name]:
                        logger.debug(
                            "Package requires %s, incompatible with policy %s which requires %s",
                            symbol,
                            policy_name,
                            policy_sym_vers[name],
                        )
                    policy_satisfied = False
            return policy_satisfied

        required_vers: dict[str, set[str]] = {}
        for symbols in versioned_symbols.values():
            for symbol in symbols:
                sym_name, _, _ = symbol.partition("_")
                required_vers.setdefault(sym_name, set()).add(symbol)
        for p in self._policies[::-1]:
            policy_sym_vers = {
                sym_name: {sym_name + "_" + version for version in versions}
                for sym_name, versions in p.symbol_versions.items()
            }
            if policy_is_satisfied(p.name, policy_sym_vers):
                return p

        # the base policy (generic linux) should always match
        msg = "Internal error"  # pragma: no cover
        raise RuntimeError(msg)  # pragma: no cover

    def lddtree_external_references(
        self,
        lddtree: DynamicExecutable,
        wheel_path: Path,
    ) -> dict[str, ExternalReference]:
        def filter_libs(
            libs: Iterable[str],
            whitelist: frozenset[str],
        ) -> Generator[str]:
            for lib in libs:
                if "ld-linux" in lib or lib in ["ld64.so.2", "ld64.so.1"]:
                    # always exclude ELF dynamic linker/loader
                    # 'ld64.so.2' on s390x
                    # 'ld64.so.1' on ppc64le
                    # 'ld-linux*' on other platforms
                    continue
                if self.libc == Libc.ANDROID and LIBPYTHON_RE.match(lib):
                    # libpython dependencies are normal on Android.
                    continue
                if lib in whitelist:
                    # exclude any libs in the whitelist
                    continue
                yield lib

        def get_req_external(libs: set[str], whitelist: frozenset[str]) -> set[str]:
            # get all the required external libraries
            libs = libs.copy()
            reqs = set()
            while libs:
                lib = libs.pop()
                reqs.add(lib)
                for dep in filter_libs(lddtree.libraries[lib].needed, whitelist):
                    if dep not in reqs:
                        libs.add(dep)
            return reqs

        ret: dict[str, ExternalReference] = {}
        for p in self._policies:
            needed_external_libs: set[str] = set()
            blacklist = {}

            if p != self.linux:
                # special-case the generic linux platform here, because it
                # doesn't have a whitelist. or, you could say its
                # whitelist is the complete set of all libraries. so nothing
                # is considered "external" that needs to be copied in.
                whitelist = p.whitelist
                blacklist_libs = set(p.blacklist.keys()) & set(lddtree.needed)
                blacklist_reduced = {k: p.blacklist[k] for k in blacklist_libs}
                blacklist = filter_undefined_symbols(
                    lddtree.realpath,
                    blacklist_reduced,
                )
                needed_external_libs = get_req_external(
                    set(filter_libs(lddtree.needed, whitelist)),
                    whitelist,
                )

            pol_ext_deps: dict[str, Path | None] = {}
            for lib in needed_external_libs:
                if is_subdir(lddtree.libraries[lib].realpath, wheel_path):
                    # we didn't filter libs that resolved via RPATH out
                    # earlier because we wanted to make sure to pick up
                    # our elf's indirect dependencies. But now we want to
                    # filter these ones out, since they're not "external".
                    logger.debug("RPATH FTW: %s", lib)
                    continue
                pol_ext_deps[lib] = lddtree.libraries[lib].realpath
            ret[p.name] = ExternalReference(pol_ext_deps, blacklist, p)
        return ret


def _validate_pep600_compliance(policies: Any) -> None:  # noqa: ANN401
    symbol_versions: dict[str, dict[str, set[str]]] = {}
    lib_whitelist: set[str] = set()
    for policy in sorted(policies, key=lambda x: x["priority"], reverse=True):
        if policy["name"] == "linux":
            continue
        if not lib_whitelist.issubset(set(policy["lib_whitelist"])):
            diff = lib_whitelist - set(policy["lib_whitelist"])
            msg = (
                'Invalid "policy.json" file. Missing whitelist libraries in '
                f'"{policy["name"]}" compared to previous policies: {diff}'
            )
            raise ValueError(msg)
        lib_whitelist.update(policy["lib_whitelist"])
        for arch in policy["symbol_versions"]:
            symbol_versions_arch = symbol_versions.get(arch, defaultdict(set))
            for prefix in policy["symbol_versions"][arch]:
                policy_symbol_versions = set(policy["symbol_versions"][arch][prefix])
                if not symbol_versions_arch[prefix].issubset(policy_symbol_versions):
                    diff = symbol_versions_arch[prefix] - policy_symbol_versions
                    msg = (
                        'Invalid "policy.json" file. Symbol versions missing '
                        f'in "{policy["name"]}_{arch}" for "{prefix}" '
                        f"compared to previous policies: {diff}"
                    )
                    raise ValueError(msg)
                symbol_versions_arch[prefix].update(
                    policy["symbol_versions"][arch][prefix],
                )
            symbol_versions[arch] = symbol_versions_arch


def _fixup_musl_libc_soname(
    libc: Libc,
    arch: Architecture,
    whitelist: list[str],
) -> frozenset[str]:
    if libc != Libc.MUSL:
        return frozenset(whitelist)
    soname_map = {
        "libc.so": {
            Architecture.x86_64: "libc.musl-x86_64.so.1",
            Architecture.i686: "libc.musl-x86.so.1",
            Architecture.aarch64: "libc.musl-aarch64.so.1",
            Architecture.s390x: "libc.musl-s390x.so.1",
            Architecture.ppc64le: "libc.musl-ppc64le.so.1",
            Architecture.armv7l: "libc.musl-armv7.so.1",
            Architecture.riscv64: "libc.musl-riscv64.so.1",
            Architecture.loongarch64: "libc.musl-loongarch64.so.1",
        },
    }
    new_whitelist = []
    for soname in whitelist:
        if soname in soname_map:
            new_soname = soname_map[soname][arch.baseline]
            logger.debug("Replacing whitelisted '%s' by '%s'", soname, new_soname)
            new_whitelist.append(new_soname)
        else:
            new_whitelist.append(soname)
    return frozenset(new_whitelist)


def get_replace_platforms(name: str) -> list[str]:
    """Extract platform tag replacement rules from policy

    >>> get_replace_platforms('linux_x86_64')
    []
    >>> get_replace_platforms('linux_i686')
    []
    >>> get_replace_platforms('manylinux1_x86_64')
    ['linux_x86_64']
    >>> get_replace_platforms('manylinux1_i686')
    ['linux_i686']

    """
    if name.startswith("linux"):
        return []
    if name.startswith("manylinux_"):
        return ["linux_" + "_".join(name.split("_")[3:])]
    if name.startswith("musllinux_"):
        return ["linux_" + "_".join(name.split("_")[3:])]
    return ["linux_" + "_".join(name.split("_")[1:])]


__all__ = [
    "WheelPolicies",
]



# Copyright 2016 Robert McGibbon
# Copyright 2012-2014 Gentoo Foundation
# Copyright 2012-2014 Mike Frysinger <vapier@gentoo.org>
# Copyright 2012-2014 The Chromium OS Authors
# Use of this source code is governed by a BSD-style license (BSD-3)
# Original version available from:
#   https://sources.gentoo.org/cgi-bin/viewvc.cgi/gentoo-projects/pax-utils/lddtree.py
"""Read the ELF dependency tree

This does not work like `ldd` in that we do not execute/load code (only read
files on disk).
"""

from __future__ import annotations

import errno
import functools
import glob
import logging
import os
import re
from dataclasses import dataclass
from fnmatch import fnmatch
from pathlib import Path

from elftools.elf.constants import E_FLAGS
from elftools.elf.elffile import ELFFile
from elftools.elf.sections import NoteSection

from auditwheel.architecture import Architecture
from auditwheel.error import InvalidLibcError
from auditwheel.libc import Libc

log = logging.getLogger(__name__)
__all__ = ["LIBPYTHON_RE", "DynamicExecutable", "DynamicLibrary", "ldd"]

# Regex to match libpython shared library names
LIBPYTHON_RE = re.compile(r"^libpython\d+\.\d+m?.so(\.\d)*$")

# Regex to match ORIGIN references in rpaths.
ORIGIN_RE = re.compile(r"\$(ORIGIN|\{ORIGIN\})")


@dataclass(frozen=True)
class Platform:
    _elf_osabi: str
    _elf_class: int
    _elf_little_endian: bool
    _elf_machine: str
    _base_arch: Architecture | None
    _ext_arch: Architecture | None
    _error_msg: str | None

    def is_compatible(self, other: Platform) -> bool:
        os_abis = frozenset((self._elf_osabi, other._elf_osabi))
        compat_sets = (frozenset(f"ELFOSABI_{x}" for x in ("NONE", "SYSV", "GNU", "LINUX")),)
        return (
            (len(os_abis) == 1 or any(os_abis.issubset(x) for x in compat_sets))
            and self._elf_class == other._elf_class
            and self._elf_little_endian == other._elf_little_endian
            and self._elf_machine == other._elf_machine
        )

    @property
    def baseline_architecture(self) -> Architecture:
        if self._base_arch is not None:
            return self._base_arch
        raise ValueError(self._error_msg)

    @property
    def extended_architecture(self) -> Architecture | None:
        if self._error_msg is not None:
            raise ValueError(self._error_msg)
        return self._ext_arch


@dataclass(frozen=True)
class DynamicLibrary:
    soname: str
    path: str | None
    realpath: Path | None
    platform: Platform | None = None
    needed: tuple[str, ...] = ()


@dataclass(frozen=True)
class DynamicExecutable:
    interpreter: str | None
    libc: Libc | None
    path: str
    realpath: Path
    platform: Platform
    needed: tuple[str, ...]
    rpath: tuple[str, ...]
    runpath: tuple[str, ...]
    libraries: dict[str, DynamicLibrary]


def _get_platform(elf: ELFFile) -> Platform:
    elf_osabi = elf.header["e_ident"]["EI_OSABI"]
    elf_class = elf.elfclass
    elf_little_endian = elf.little_endian
    elf_machine = elf["e_machine"]
    base_arch = {
        ("EM_386", 32, True): Architecture.i686,
        ("EM_X86_64", 64, True): Architecture.x86_64,
        ("EM_PPC64", 64, True): Architecture.ppc64le,
        ("EM_PPC64", 64, False): Architecture.ppc64,
        ("EM_RISCV", 64, True): Architecture.riscv64,
        ("EM_AARCH64", 64, True): Architecture.aarch64,
        ("EM_S390", 64, False): Architecture.s390x,
        ("EM_ARM", 32, True): Architecture.armv7l,
        ("EM_LOONGARCH", 64, True): Architecture.loongarch64,
    }.get((elf_machine, elf_class, elf_little_endian))
    ext_arch: Architecture | None = None
    error_msg: str | None = None
    flags = elf["e_flags"]
    assert base_arch is None or base_arch.baseline == base_arch  # noqa: S101
    if base_arch is None:
        error_msg = "Unknown architecture"
    elif base_arch == Architecture.x86_64:
        for section in elf.iter_sections():
            if not isinstance(section, NoteSection):
                continue
            for note in section.iter_notes():
                if note["n_type"] != "NT_GNU_PROPERTY_TYPE_0":
                    continue
                if note["n_name"] != "GNU":
                    continue
                for prop in note["n_desc"]:
                    if prop.pr_type != "GNU_PROPERTY_X86_ISA_1_NEEDED":
                        continue
                    if prop.pr_datasz != 4:
                        continue
                    data = prop.pr_data
                    data -= data & 1  # clear baseline
                    if data & 8 == 8:
                        ext_arch = Architecture.x86_64_v4
                        break
                    if data & 4 == 4:
                        ext_arch = Architecture.x86_64_v3
                        break
                    if data & 2 == 2:
                        ext_arch = Architecture.x86_64_v2
                        break
                    if data != 0:
                        error_msg = "unknown x86_64 ISA"
                break
    elif base_arch == Architecture.armv7l:
        if (flags & E_FLAGS.EF_ARM_EABIMASK) != E_FLAGS.EF_ARM_EABI_VER5:
            error_msg = "Invalid ARM EABI version for armv7l"
        elif (flags & E_FLAGS.EF_ARM_ABI_FLOAT_HARD) != E_FLAGS.EF_ARM_ABI_FLOAT_HARD:
            error_msg = "armv7l shall use hard-float"
        if error_msg is not None:
            base_arch = None

    return Platform(
        elf_osabi,
        elf_class,
        elf_little_endian,
        elf_machine,
        base_arch,
        ext_arch,
        error_msg,
    )


def normpath(path: str) -> str:
    """Normalize a path

    Python's os.path.normpath() doesn't handle some cases:
    // -> //
    //..// -> //
    //..//..// -> ///
    """
    return os.path.normpath(path).replace("//", "/")


def readlink(path: str, root: str, *, prefixed: bool = False) -> str:
    """Like os.readlink(), but relative to a ``root``

    This does not currently handle the pathological case:
    /lib/foo.so -> ../../../../../../../foo.so
    This relies on the .. entries in / to point to itself.

    Parameters
    ----------
    path
        The symlink to read
    root
        The path to use for resolving absolute symlinks
    prefixed
        When False, the ``path`` must not have ``root`` prefixed to it, nor
        will the return value have ``root`` prefixed.  When True, ``path``
        must have ``root`` prefixed, and the return value will have ``root``
        added.

    Returns
    -------
    A fully resolved symlink path

    """
    root = root.rstrip("/")
    if prefixed:
        path = path[len(root) :]

    while os.path.islink(root + path):
        path = os.path.join(os.path.dirname(path), os.readlink(root + path))

    return normpath((root + path) if prefixed else path)


def dedupe(items: list[str]) -> list[str]:
    """Remove all duplicates from ``items`` (keeping order)"""
    seen: dict[str, str] = {}
    return [seen.setdefault(x, x) for x in items if x not in seen]


def parse_ld_paths(str_ldpaths: str, path: str = "", root: str = "") -> list[str]:
    """Parse the colon-delimited list of paths and apply ldso rules to each

    Note the special handling as dictated by the ldso:
    - Empty paths are equivalent to $PWD
    - $ORIGIN is expanded to the directory containing the given file
    - (TODO) $LIB and friends

    Parameters
    ----------
    str_ldpaths
        A colon-delimited string of paths
    root
        The path to prepend to all paths found
    path
        The object actively being parsed (used for $ORIGIN)

    Returns
    -------
        list of processed paths

    """
    if not str_ldpaths:
        return []

    ldpaths: list[str] = []
    for ldpath in str_ldpaths.split(":"):
        if ldpath == "":
            # The ldso treats "" paths as $PWD.
            ldpath_ = os.getcwd()
        elif re.search(ORIGIN_RE, ldpath):
            if not path:
                msg = "can't expand $ORIGIN without a path"
                raise ValueError(msg)
            ldpath_ = re.sub(ORIGIN_RE, os.path.dirname(os.path.abspath(path)), ldpath)
        else:
            ldpath_ = root + ldpath
        ldpaths.append(normpath(ldpath_))
    return [p for p in dedupe(ldpaths) if os.path.isdir(p)]


@functools.lru_cache
def parse_ld_so_conf(ldso_conf: str, *, root: str = "/", _first: bool = True) -> list[str]:
    """Load all the paths from a given ldso config file

    This should handle comments, whitespace, and "include" statements.

    Parameters
    ----------
    ldso_conf
        The file to scan
    root
        The path to prepend to all paths found
    _first
        Recursive use only; is this the first ELF?

    Returns
    -------
    list of paths found

    """
    paths: list[str] = []

    dbg_pfx = "" if _first else "  "
    try:
        log.debug("%sparse_ld_so_conf(%s)", dbg_pfx, ldso_conf)
        with open(ldso_conf) as f:
            for input_line in f:
                line = input_line.split("#", 1)[0].strip()
                if not line:
                    continue
                if line.startswith("include "):
                    line = line[8:]
                    if line[0] == "/":
                        line = root + line.lstrip("/")
                    else:
                        line = os.path.dirname(ldso_conf) + "/" + line
                    log.debug("%s  glob: %s", dbg_pfx, line)
                    for path in glob.glob(line):
                        paths += parse_ld_so_conf(path, root=root, _first=False)
                else:
                    paths += [normpath(root + line)]
    except OSError as e:
        if e.errno != errno.ENOENT:
            log.warning(e)

    if _first:
        # TODO: Load paths from ldso itself.
        # Remove duplicate entries to speed things up.
        paths = [p for p in dedupe(paths) if os.path.isdir(p)]

    return paths


@functools.lru_cache
def load_ld_paths(
    libc: Libc | None,
    root: str = "/",
    prefix: str = "",
) -> dict[str, list[str]]:
    """Load linker paths from common locations

    This parses the ld.so.conf and LD_LIBRARY_PATH env var.

    Parameters
    ----------
    root
        The root tree to prepend to paths
    prefix
        The path under ``root`` to search

    Returns
    -------
    dict containing library paths to search

    """
    ldpaths: dict[str, list[str]] = {"conf": [], "env": [], "interp": []}

    # Load up $LD_LIBRARY_PATH.
    env_ldpath = os.environ.get("LD_LIBRARY_PATH")
    if env_ldpath is not None:
        if root != "/":
            log.warning("ignoring LD_LIBRARY_PATH due to ROOT usage")
        else:
            # TODO: If this contains $ORIGIN, we probably have to parse this
            # on a per-ELF basis so it can get turned into the right thing.
            ldpaths["env"] = parse_ld_paths(env_ldpath, path="")

    if libc == Libc.MUSL:
        # from https://git.musl-libc.org/cgit/musl/tree/ldso
        # /dynlink.c?id=3f701faace7addc75d16dea8a6cd769fa5b3f260#n1063
        root_prefix = Path(root) / prefix
        ld_musl = list((root_prefix / "etc").glob("ld-musl-*.path"))
        assert len(ld_musl) <= 1  # noqa: S101
        if len(ld_musl) == 0:
            ldpaths["conf"] = [
                root + "/lib",
                root + "/usr/local/lib",
                root + "/usr/lib",
            ]
        else:
            ldpaths["conf"] = []
            for ldpath in ld_musl[0].read_text().split(":"):
                ldpath_stripped = ldpath.strip()
                if ldpath_stripped == "":
                    continue
                ldpaths["conf"].append(root + ldpath_stripped)
    else:
        # Load up /etc/ld.so.conf.
        ldpaths["conf"] = parse_ld_so_conf(root + prefix + "/etc/ld.so.conf", root=root)
        # the trusted directories are not necessarily in ld.so.conf
        ldpaths["conf"].extend(["/lib", "/lib64/", "/usr/lib", "/usr/lib64"])
    log.debug("linker ldpaths: %s", ldpaths)
    return ldpaths


def ld_paths_from_arg(args_ldpaths: str | None) -> dict[str, list[str]] | None:
    """Load linker paths from the --ldpaths option and the environment."""
    if args_ldpaths is None:
        # The option was not provided, so fall back on load_ld_paths.
        return None

    return {
        "conf": parse_ld_paths(args_ldpaths),
        "env": parse_ld_paths(os.environ.get("AUDITWHEEL_LD_LIBRARY_PATH", "")),
        "interp": [],
    }


def find_lib(
    platform: Platform,
    lib: str,
    ldpaths: list[str],
    root: str = "/",
) -> tuple[Path | None, str | None]:
    """Try to locate a ``lib`` that is compatible to ``elf`` in the given
    ``ldpaths``

    Parameters
    ----------
    platform : Platform
        The platform which the library should be compatible with (ELF wise)
    lib : str
        The library (basename) to search for
    ldpaths : list[str]
        A list of paths to search
    root : str
       The root path to resolve symlinks

    Returns
    -------
    Tuple of the full path to the desired library and the real path to it

    """
    for ldpath in ldpaths:
        path = os.path.join(ldpath, lib)
        target = Path(readlink(path, root, prefixed=True))

        if target.exists():
            with open(target, "rb") as f:
                libelf = ELFFile(f)
                if platform.is_compatible(_get_platform(libelf)):
                    return target, path

    return None, None


def ldd(
    path: Path,
    root: str = "/",
    prefix: str = "",
    ldpaths: dict[str, list[str]] | None = None,
    display: str | None = None,
    exclude: frozenset[str] = frozenset(),
    _all_libs: dict[str, DynamicLibrary] | None = None,
) -> DynamicExecutable:
    """Parse the ELF dependency tree of the specified file

    Parameters
    ----------
    path
        The ELF to scan
    root
        The root tree to prepend to paths; this applies to interp and rpaths
        only as ``path`` and ``ldpaths`` are expected to be prefixed already
    prefix
        The path under ``root`` to search
    ldpaths
        dict containing library paths to search; should have the keys:
        conf, env, interp. If not supplied, the function ``load_ld_paths``
        will be called.
    display
        The path to show rather than ``path``
    exclude
        List of soname (DT_NEEDED) to exclude from the tree
    _all_libs
        Recursive use only; dict of all libs we've seen

    Returns
    -------
    a dict containing information about all the ELFs; e.g.
    {
      'interp': '/lib64/ld-linux.so.2',
      'needed': ['libc.so.6', 'libcurl.so.4',],
      'libs': {
        'libc.so.6': {
          'path': '/lib64/libc.so.6',
          'needed': [],
        },
        'libcurl.so.4': {
          'path': '/usr/lib64/libcurl.so.4',
          'needed': ['libc.so.6', 'librt.so.1',],
        },
      },
    }

    """
    _first = _all_libs is None
    if _all_libs is None:
        _all_libs = {}

    log.debug("ldd(%s)", path)

    interpreter: str | None = None
    libc: Libc | None = None
    needed: list[str] = []
    rpaths: list[str] = []
    runpaths: list[str] = []

    with open(path, "rb") as f:
        elf = ELFFile(f)

        # get the platform
        platform = _get_platform(elf)

        # If this is the first ELF, extract the interpreter.
        if _first:
            for segment in elf.iter_segments():
                if segment.header.p_type != "PT_INTERP":
                    continue
                interp = segment.get_interp_name()
                log.debug("  interp           = %s", interp)
                interpreter = normpath(root + interp)
                soname = os.path.basename(interpreter)
                _all_libs[soname] = DynamicLibrary(
                    soname,
                    interpreter,
                    Path(readlink(interpreter, root, prefixed=True)),
                    platform,
                )
                # if we have an interpreter and it's not MUSL, assume GLIBC
                libc = Libc.MUSL if soname.startswith("ld-musl-") else Libc.GLIBC
                if ldpaths is None:
                    ldpaths = load_ld_paths(libc).copy()
                # TODO: Should read it and scan for /lib paths.
                ldpaths["interp"] = [
                    normpath(root + os.path.dirname(interp)),
                    normpath(
                        root + prefix + "/usr" + os.path.dirname(interp).lstrip(prefix),
                    ),
                ]
                log.debug("  ldpaths[interp]  = %s", ldpaths["interp"])
                break

        # Parse the ELF's dynamic tags.
        for segment in elf.iter_segments():
            if segment.header.p_type != "PT_DYNAMIC":
                continue
            for t in segment.iter_tags():
                if t.entry.d_tag == "DT_RPATH":
                    rpaths = parse_ld_paths(t.rpath, path=str(path), root=root)
                elif t.entry.d_tag == "DT_RUNPATH":
                    runpaths = parse_ld_paths(t.runpath, path=str(path), root=root)
                elif t.entry.d_tag == "DT_NEEDED":
                    needed.append(t.needed)
            if runpaths:
                # If both RPATH and RUNPATH are set, only the latter is used.
                rpaths = []

            # We assume there is only one PT_DYNAMIC.  This is
            # probably fine since the runtime ldso does the same.
            break

        del elf

    if _first:
        # get the libc based on dependencies
        for soname in needed:
            if soname.startswith(("libc.musl-", "ld-musl-")):
                if libc is None:
                    libc = Libc.MUSL
                if libc != Libc.MUSL:
                    msg = f"found a dependency on MUSL but the libc is already set to {libc}"
                    raise InvalidLibcError(msg)
            elif soname == "libc.so.6" or soname.startswith(("ld-linux-", "ld64.so.")):
                if libc is None:
                    libc = Libc.GLIBC
                if libc != Libc.GLIBC:
                    msg = f"found a dependency on GLIBC but the libc is already set to {libc}"
                    raise InvalidLibcError(msg)
        if libc is None:
            # try the filename as a last resort
            if path.name.endswith(("-arm-linux-musleabihf.so", "-linux-musl.so")):
                libc = Libc.MUSL
            elif path.name.endswith(("-arm-linux-gnueabihf.so", "-linux-gnu.so")):
                # before python 3.11, musl was also using gnu
                soabi = path.stem.split(".")[-1].split("-")
                valid_python = tuple(f"3{minor}" for minor in range(11, 100))
                if soabi[0] == "cpython" and soabi[1].startswith(valid_python):
                    libc = Libc.GLIBC

        if ldpaths is None:
            ldpaths = load_ld_paths(libc).copy()
        # Propagate the rpaths used by the main ELF since those will be
        # used at runtime to locate things.
        ldpaths["rpath"] = rpaths
        ldpaths["runpath"] = runpaths
        log.debug("  ldpaths[rpath]   = %s", rpaths)
        log.debug("  ldpaths[runpath] = %s", runpaths)

    assert ldpaths is not None  # noqa: S101

    all_ldpaths = (
        ldpaths["rpath"]
        + rpaths
        + runpaths
        + ldpaths["env"]
        + ldpaths["runpath"]
        + ldpaths["conf"]
        + ldpaths["interp"]
    )
    _excluded_libs: set[str] = set()
    for soname in needed:
        if soname in _all_libs:
            continue
        if soname in _excluded_libs:
            continue
        if any(fnmatch(soname, e) for e in exclude):
            log.info("Excluding %s", soname)
            _excluded_libs.add(soname)
            continue

        # special case for libpython, see https://github.com/pypa/auditwheel/issues/589
        # we want to return the dependency to be able to remove it later on but
        # we don't want to analyze it for symbol versions nor do we want to analyze its
        # dependencies as it will be removed.
        if LIBPYTHON_RE.match(soname):
            log.info("Skip %s resolution", soname)
            _all_libs[soname] = DynamicLibrary(soname, None, None)
            continue

        realpath, fullpath = find_lib(platform, soname, all_ldpaths, root)
        if realpath is not None and any(fnmatch(str(realpath), e) for e in exclude):
            log.info("Excluding %s", realpath)
            _excluded_libs.add(soname)
            continue
        _all_libs[soname] = DynamicLibrary(soname, fullpath, realpath)
        if realpath is None or fullpath is None:
            continue
        dependency = ldd(realpath, root, prefix, ldpaths, fullpath, exclude, _all_libs)
        _all_libs[soname] = DynamicLibrary(
            soname,
            fullpath,
            realpath,
            dependency.platform,
            dependency.needed,
        )

    return DynamicExecutable(
        interpreter,
        libc,
        str(path) if display is None else display,
        path,
        platform,
        tuple(soname for soname in needed if soname not in _excluded_libs),
        tuple(rpaths),
        tuple(runpaths),
        _all_libs,
    )



"""Options shared between multiple subcommands."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from argparse import ArgumentParser


def disable_isa_check(parser: ArgumentParser) -> None:
    parser.add_argument(
        "--disable-isa-ext-check",
        dest="DISABLE_ISA_EXT_CHECK",
        action="store_true",
        help="Do not check for extended ISA compatibility (e.g. x86_64_v2)",
        default=False,
    )


def allow_pure_python_wheel(parser: ArgumentParser) -> None:
    parser.add_argument(
        "--allow-pure-python-wheel",
        dest="ALLOW_PURE_PY_WHEEL",
        action="store_true",
        help="Allow processing of pure Python wheels (no platform-specific binaries) without error",
        default=False,
    )


def ldpaths(parser: ArgumentParser) -> None:
    parser.add_argument(
        "--ldpaths",
        dest="LDPATHS",
        help="Colon-delimited list of directories to search for external libraries. "
        "This replaces the default list; to add to this list, set the environment "
        "variable AUDITWHEEL_LD_LIBRARY_PATH.",
    )



"""Contexts for *with* statement providing temporary directories"""

from __future__ import annotations

import os
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from types import TracebackType


class InTemporaryDirectory:
    """Create, return, and change directory to a temporary directory

    Examples
    --------
    >>> from pathlib import Path
    >>> my_cwd = Path.cwd()
    >>> with InTemporaryDirectory() as tmpdir:
    ...     _ = open('test.txt', 'wt').write('some text')
    ...     assert os.path.isfile('test.txt')
    ...     assert tmpdir.joinpath('test.txt').is_file()
    >>> tmpdir.exists()
    False
    >>> Path.cwd() == my_cwd
    True

    """

    def __init__(self) -> None:
        self._tmpdir = TemporaryDirectory()
        self._name = Path(self._tmpdir.name).resolve(strict=True)

    @property
    def name(self) -> Path:
        return self._name

    def __enter__(self) -> Path:
        self._pwd = Path.cwd()
        os.chdir(self._name)
        self._tmpdir.__enter__()
        return self._name

    def __exit__(
        self,
        exc: type[BaseException] | None,
        value: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        os.chdir(self._pwd)
        self._tmpdir.__exit__(exc, value, tb)


class InGivenDirectory:
    """Change directory to given directory for duration of ``with`` block

    Useful when you want to use `InTemporaryDirectory` for the final test, but
    you are still debugging.  For example, you may want to do this in the end:

    >>> with InTemporaryDirectory() as tmpdir:
    ...     # do something complicated which might break
    ...     pass

    But indeed the complicated thing does break, and meanwhile the
    ``InTemporaryDirectory`` context manager wiped out the directory with the
    temporary files that you wanted for debugging.  So, while debugging, you
    replace with something like:

    >>> with InGivenDirectory() as tmpdir: # Use working directory by default
    ...     # do something complicated which might break
    ...     pass

    You can then look at the temporary file outputs to debug what is happening,
    fix, and finally replace ``InGivenDirectory`` with ``InTemporaryDirectory``
    again.
    """

    def __init__(self, path: Path | None = None) -> None:
        """Initialize directory context manager

        Parameters
        ----------
        path : None or Path, optional
            path to change directory to, for duration of ``with`` block.
            Defaults to ``Path.cwd()`` if None

        """
        if path is None:
            path = Path.cwd()
        self.name = path.absolute()

    def __enter__(self) -> Path:
        self._pwd = Path.cwd()
        if not self.name.is_dir():
            self.name.mkdir()
        os.chdir(self.name)
        return self.name

    def __exit__(
        self,
        exc: type[BaseException] | None,
        value: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        os.chdir(self._pwd)



from __future__ import annotations

import os
import subprocess
import sysconfig
from importlib.metadata import distribution


def runit(x):
    dist = distribution("testpackage")
    scripts_paths = [
        os.path.abspath(sysconfig.get_path("scripts", scheme))
        for scheme in sysconfig.get_scheme_names()
    ]
    scripts = []
    for file in dist.files:
        if os.path.abspath(str(file.locate().parent)) in scripts_paths:
            scripts.append(file.locate().resolve(strict=True))
    assert len(scripts) == 2, scripts
    filename = next(script for script in scripts if script.stem == "testprogram")
    output = subprocess.check_output([filename, str(x)])
    return float(output)



"""Context managers like those in wheeltools.py for unpacking
conda packages.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from auditwheel.tmpdirs import InTemporaryDirectory
from auditwheel.tools import tarbz2todir

if TYPE_CHECKING:
    from pathlib import Path


class InCondaPkg(InTemporaryDirectory):
    def __init__(self, in_conda_pkg: Path) -> None:
        """Initialize in-conda-package context manager"""
        self.in_conda_pkg = in_conda_pkg.absolute()
        super().__init__()

    def __enter__(self) -> Path:
        tarbz2todir(self.in_conda_pkg, self.name)
        return super().__enter__()


class InCondaPkgCtx(InCondaPkg):
    def __init__(self, in_conda_pkg: Path) -> None:
        super().__init__(in_conda_pkg)
        self.path: Path | None = None

    def __enter__(self):  # type: ignore[no-untyped-def]
        self.path = super().__enter__()
        return self

    def iter_files(self) -> list[str]:
        if self.path is None:
            msg = "This function should be called from context manager"
            raise ValueError(msg)
        files = self.path / "info" / "files"
        with files.open() as f:
            return [line.strip() for line in f.readlines()]



import os
import subprocess
from pathlib import Path
from zipfile import ZipFile

import pytest
from elftools.elf.elffile import ELFFile

from auditwheel.architecture import Architecture
from auditwheel.elfutils import elf_read_dt_needed
from auditwheel.libc import Libc
from auditwheel.policy import ExternalReference
from auditwheel.wheel_abi import analyze_wheel_abi

HERE = Path(__file__).parent.resolve()
android_dir = HERE / "android"

# This wheel was generated from cibuildwheel's test_android.py::test_libcxx. It contains an
# external reference to libc++_shared.so.
libcxx_wheel = android_dir / "spam-0.1.0-cp313-cp313-android_24_arm64_v8a.whl"
libcxx_module = "spam.cpython-313-aarch64-linux-android.so"


def elf_read_tag(fn: Path, tag: str) -> str | None:
    with fn.open("rb") as f:
        elf = ELFFile(f)
        section = elf.get_section_by_name(".dynamic")
        if section:
            for t in section.iter_tags():
                if t.entry.d_tag == f"DT_{tag.upper()}":
                    return str(getattr(t, tag))
        return None


@pytest.mark.parametrize("ldpaths_methods", [["env"], ["arg"], ["env", "arg"]])
def test_libcxx(ldpaths_methods, tmp_path):
    wheelhouse = tmp_path / "wheelhouse"
    wheelhouse.mkdir()

    # The dummy libc++_shared.so files were generated using this command, where SUBDIR is either
    # "env" or "arg":
    #     echo "void SUBDIR() {}"
    #     | aarch64-linux-android24-clang -x c - -shared -o SUBDIR/libc++_shared.so
    ldpaths_args: list[str | Path] = []
    if "arg" in ldpaths_methods:
        libcxx_hash = "1be9716c"
        ldpaths_args = ["--ldpaths", android_dir / "ldpaths/arg"]

    env = os.environ.copy()
    if "env" in ldpaths_methods:
        # AUDITWHEEL_LD_LIBRARY_PATH is searched before --ldpaths, so we overwrite the hash.
        libcxx_hash = "18b6a03d"
        env["AUDITWHEEL_LD_LIBRARY_PATH"] = str(android_dir / "ldpaths/env")

    subprocess.run(
        ["auditwheel", "repair", "-w", wheelhouse, *ldpaths_args, libcxx_wheel],
        env=env,
        text=True,
        check=True,
    )
    output_wheels = list(wheelhouse.iterdir())
    assert len(output_wheels) == 1
    assert output_wheels[0].name == libcxx_wheel.name

    output_dir = tmp_path / "output"
    output_dir.mkdir()
    with ZipFile(output_wheels[0]) as zf:
        zf.extractall(output_dir)

    libs_dir = output_dir / "spam.libs"
    libcxx_path = libs_dir / f"libc++_shared-{libcxx_hash}.so"
    assert elf_read_tag(libcxx_path, "soname") == libcxx_path.name
    assert elf_read_tag(libcxx_path, "rpath") is None
    assert elf_read_tag(libcxx_path, "runpath") is None

    spam_path = output_dir / libcxx_module
    assert set(elf_read_dt_needed(spam_path)) == {
        # Included in the policy
        "libc.so",
        "libdl.so",
        "libm.so",
        # libpython dependency is normal on Android, so it should be left alone
        "libpython3.13.so",
        # Grafted library
        libcxx_path.name,
    }
    assert elf_read_tag(spam_path, "rpath") is None
    assert elf_read_tag(spam_path, "runpath") == f"$ORIGIN/{libs_dir.name}"


def test_analyze_wheel_abi():
    winfo = analyze_wheel_abi(
        Libc.ANDROID,
        Architecture.aarch64,
        libcxx_wheel,
        exclude=frozenset(),
        disable_isa_ext_check=False,
        allow_graft=True,
    )
    policy_name = "android_24_arm64_v8a"
    assert winfo.overall_policy.name == policy_name
    external_ref = ExternalReference(
        libs={"libc++_shared.so": None},
        blacklist={},
        policy=winfo.overall_policy,
    )
    assert winfo.external_refs[policy_name] == external_ref
    assert winfo.full_external_refs[Path(libcxx_module)] == winfo.external_refs
